# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Outer runner to manage interacting HTML outputs in IPython notebooks.

When rendering HTML outputs in IPython notebooks, we want to support:

- Compression, so that large repetitive HTML outputs do not cause excessively
  large notebook sizes,
- Streaming output, where the structure of an output can be sent to the
  browser while NDArrays and other contents are still being loaded and
  rendered in the IPython kernel,
- Final output stealing (for streaming output mode), where the temporary
  rendering can be moved into the final "result" display cell in IPython
  systems that show Out[...] markers,
- Duplication-safety, so that if the same output is repeated multiple times in
  a notebook display environment (e.g. in a separate view-only cell), each copy
  is rendered independently,
- Shadow-root encapsulation, where the rendering is contained in the "shadow
  DOM". This keeps its JS/CSS separate from the main notebook, prevents
  interference between different renderings, and may allow rendering
  optimizations in some browsers.

This module defines the necessary logic to enable these features. It takes as
input a sequence of HTML blobs that should be rendered into a single container
in a streaming fashion, and transforms them into a new sequence of compressed,
duplication-safe HTML blobs that first create the container and then render each
of the contents into it.
"""

from __future__ import annotations

import base64
import dataclasses
import enum
import io
from typing import Iterable, Iterator
import uuid
import zlib


class SegmentType(enum.Enum):
  """Type of HTML segment."""

  # The blob that creates and owns the container.
  CONTAINER = enum.auto()
  # A blob that inserts new data into the initial container, but doesn't display
  # any content itself.
  CONTAINER_UPDATE = enum.auto()
  # A final blob that will take ownership of the container, moving the container
  # into itself. Only used when stealing is enabled.
  FINAL_OUTPUT_STEALER = enum.auto()


@dataclasses.dataclass(frozen=True)
class HTMLOutputSegment:
  """A piece of partial HTML output."""

  html_src: str
  segment_type: SegmentType


CONTAINER_TEMPLATE = """
<script>
(()=>{
  if (customElements.get('treescope-container') === undefined) {
    // Custom container element that contains the rendering and also owns
    // JavaScript definitions for the rendering. Intentionally does not contain
    // fixed implementations of those definitions, because in some notebook
    // environments (e.g. JupyterLab) every notebook output exists in the global
    // browser window namespace, and we can't have multiple versions of the
    // same custom element.
    class TreescopeContainer extends HTMLElement {
      constructor() {
        super();
        this.attachShadow({mode: "open"});
        this.defns = {};
        this.state = {};
      }
    }
    customElements.define("treescope-container", TreescopeContainer);
  }
  if (customElements.get('treescope-run-here') === undefined) {
    // Custom "run-here" element that runs a script in the context of a DOM
    // node. This element should contain a single script tag with the attribute
    // type="application/octet-stream" that contains the script to run. The
    // script will be executed inside an anonymous function, with `this` bound
    // to the DOM node. Afterward, the element will be removed.
    class RunHere extends HTMLElement {
      constructor() { super() }
      connectedCallback() {
        const run = child => {
          const fn = new Function(child.textContent);
          child.textContent = "";
          fn.call(this);
          this.remove();
        };
        const child = this.querySelector("script");
        if (child) {
          run(child);
        } else {
          // In some cases, the connected callback may fire before the script
          // tag has actually been added; this can occur when loading from an
          // iframe srcdoc attribute. In this case, we need to wait for the
          // script tag to appear. We assume the only mutation that will occur
          // is adding the script tag, since the element is only supposed to
          // contain a single script tag.
          new MutationObserver(()=>{
            run(this.querySelector("script"));
          }).observe(this, {childList: true});
        }
      }
    }
    customElements.define("treescope-run-here", RunHere);
  }
})();
</script>
<treescope-container class="{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}"
></treescope-container>
<treescope-run-here><script type="application/octet-stream">
// Find the target container based on its ID class. We use a class instead of
// an HTML id because it's possible that the notebook display system will
// display this multiple times. Using the suffix, we can make sure to only
// return each container once for each suffix, so that if we duplicate the
// container and also a script that modifies it, each script will modify a
// different container.
const root = (
  Array.from(document.getElementsByClassName(
      "{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}"))
  .filter((elt) => !elt.dataset.setup)
)[0];
root.dataset.setup = 1;
const msg = document.createElement("span");
msg.style = "color: #cccccc; font-family: monospace;";
msg.textContent = "(Loading...)";
root.state.loadingMsg = msg;
root.shadowRoot.appendChild(msg);
root.state.chain = new Promise((resolve, reject) => {
  const observer = new IntersectionObserver((entries) => {
    for (const entry of entries) {
      if (entry.isIntersecting) {
        resolve();
        observer.disconnect();
        return;
      }
    }
  }, {rootMargin: "1000px"});
  window.setTimeout(() => {
    observer.observe(root);
  }, 0);
});
root.state.deferring = false;
const _insertNode = (node) => {
  // Script trick: Script tags that were inserted using innerHTML don't
  // execute when they get added to the document. The scripts in the
  // templates may have been inserted this way, which would prevent them
  // from executing in the root. To get them to execute, we need to
  // rebuild the scripts before inserting them.
  for (let oldScript of node.querySelectorAll("script")) {
      let newScript = document.createElement("script");
      newScript.type = oldScript.type;
      newScript.textContent = oldScript.textContent;
      oldScript.parentNode.replaceChild(newScript, oldScript);
  }
  if (root.state.loadingMsg) {
    root.state.loadingMsg.remove();
    root.state.loadingMsg = null;
  }
  root.shadowRoot.appendChild(node);
};
root.defns.insertContent = ((contentNode, compressed) => {
  if (compressed) {
    root.state.deferring = true;
  }
  if (root.state.deferring) {
    root.state.chain = (async () => {
      await root.state.chain;
      if (compressed) {
        // contentNode is a script.
        const encoded = contentNode.textContent;
        const blob = new Blob([
            Uint8Array.from(atob(encoded), (m) => m.codePointAt(0))
        ]);
        const reader = blob.stream().pipeThrough(
          new DecompressionStream("deflate")
        ).pipeThrough(
          new TextDecoderStream("utf-8")
        ).getReader();
        const parts = [];
        while (true) {
          const step = await reader.read();
          if (step.done) { break; }
          parts.push(step.value);
        }
        const tpl = document.createElement('template');
        tpl.innerHTML = parts.join("");
        _insertNode(tpl.content);
      } else {
        // contentNode is a template.
        _insertNode(contentNode.content);
      }
    })();
  } else {
    // contentNode is a template.
    _insertNode(contentNode.content);
  }
});
</script></treescope-run-here>
"""

STEP_TEMPLATE = """
<div style="display:none">
<template>{__REPLACE_ME_WITH_CONTENT_HTML__}</template>
<treescope-run-here><script type="application/octet-stream">
const root = (
  Array.from(document.getElementsByClassName(
      "{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}"))
  .filter((elt) => !elt.dataset['step{__REPLACE_ME_WITH_STEP__}'])
)[0];
root.dataset['step{__REPLACE_ME_WITH_STEP__}'] = 1;
root.defns.insertContent(this.parentNode.querySelector('template'), false);
this.parentNode.remove();
</script></treescope-run-here>
</div>
"""

COMPRESSED_STEP_TEMPLATE = """
<div style="display:none">
<script type="application/octet-stream"
>{__REPLACE_ME_WITH_COMPRESSED_CONTENT_HTML__}</script>
<treescope-run-here><script type="application/octet-stream">
const root = (
  Array.from(document.getElementsByClassName(
      "{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}"))
  .filter((elt) => !elt.dataset['step{__REPLACE_ME_WITH_STEP__}'])
)[0];
root.dataset['step{__REPLACE_ME_WITH_STEP__}'] = 1;
root.defns.insertContent(
    this.parentNode.querySelector('script[type="application/octet-stream"]'),
    true
);
this.parentNode.remove();
</script></treescope-run-here>
</div>
"""

STEALER_TEMPLATE = """
<treescope-run-here><script type="application/octet-stream">
const root = (
  Array.from(document.getElementsByClassName(
      "{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}"))
  .filter((elt) => !elt.dataset.stolen)
)[0];
root.dataset.stolen = 1;
this.parentNode.replaceChild(root, this);
</script></treescope-run-here>
"""


def _prep_html_js_and_strip_comments(src):
  stream = io.StringIO()
  for line in src.splitlines():
    stripped = line.strip()
    if stripped and not stripped.startswith("//"):
      stream.write(stripped)
      stream.write(" ")
  return stream.getvalue()[:-1]


def encapsulate_streaming_html(
    inner_iterator: Iterable[str],
    *,
    compress: bool = True,
    stealable: bool = False,
) -> Iterator[HTMLOutputSegment]:
  """Encapsulates a sequence of inner HTML blobs into robust iframe updates.

  This function accepts an iterator of HTML blobs that should each run in the
  same iframe, and transforms them into another iterator of HTML blobs that
  can be inserted directly into a notebook environment. The first output will
  set up the iframe, and all later updates will insert content into that
  original iframe. Optionally, the final output will be a "stealer" that will
  move the iframe into itself, ensuring that the iframe is associated with the
  correct "result" cell in IPython notebook systems that show Out[...] markers.
  Updates will be duplication-safe, in the sense that repeating the same
  sequence of outputs in a single HTML page will produce multiple copies of the
  iframe, each with the same contents, and the code in the inner iterator will
  only be executed once in each iframe.

  Args:
    inner_iterator: A nonempty iterator of HTML blobs to encapsulate.
    compress: Whether to compress the HTML blobs.
    stealable: Whether to include a final "stealer" blob that will move the
      iframe into itself.

  Yields:
    HTML output segments that can be displayed in a notebook environment or
    saved.
  """
  inner_iterator = iter(inner_iterator)

  stream = io.StringIO()
  # Build the initial iframe, and assign it a unique ID.
  # Note: This is unique in the Python program, but if the output is repeated
  # multiple times in the notebook output, we may have multiple iframes with
  # the same ID.
  unique_id_class = f"treescope_out_{uuid.uuid4().hex}"

  outer_content = _prep_html_js_and_strip_comments(CONTAINER_TEMPLATE).replace(
      "{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}", unique_id_class
  )
  stream.write(outer_content)

  for i, step_content in enumerate(inner_iterator):
    if compress:
      # Compress the input string. We use ZLIB, which is natively supported by
      # modern browsers.
      compressed = zlib.compress(
          step_content.encode("utf-8"), zlib.Z_BEST_COMPRESSION
      )
      # Serialize it as base64.
      serialized = base64.b64encode(compressed).decode("ascii")
      # Embed it.
      step_content = (
          _prep_html_js_and_strip_comments(COMPRESSED_STEP_TEMPLATE)
          .replace("{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}", unique_id_class)
          .replace("{__REPLACE_ME_WITH_STEP__}", str(i))
          .replace("{__REPLACE_ME_WITH_COMPRESSED_CONTENT_HTML__}", serialized)
      )
    else:
      step_content = (
          _prep_html_js_and_strip_comments(STEP_TEMPLATE)
          .replace("{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}", unique_id_class)
          .replace("{__REPLACE_ME_WITH_STEP__}", str(i))
          .replace("{__REPLACE_ME_WITH_CONTENT_HTML__}", step_content)
      )
    stream.write(step_content)
    if i == 0:
      segment_type = SegmentType.CONTAINER
    else:
      segment_type = SegmentType.CONTAINER_UPDATE
    yield HTMLOutputSegment(
        html_src=stream.getvalue(), segment_type=segment_type
    )
    stream = io.StringIO()

  if stealable:
    stealer_content = _prep_html_js_and_strip_comments(
        STEALER_TEMPLATE
    ).replace("{__REPLACE_ME_WITH_CONTAINER_ID_CLASS__}", unique_id_class)
    yield HTMLOutputSegment(
        html_src=stealer_content,
        segment_type=SegmentType.FINAL_OUTPUT_STEALER,
    )


def encapsulate_html(html_src: str, compress: bool = True) -> str:
  """Encapsulates HTML source code into a duplication-safe container.

  Args:
    html_src: The HTML source code to encapsulate.
    compress: Whether to compress the HTML source code.

  Returns:
    An HTML output segment that can be displayed in a notebook environment or
    saved.
  """
  [converted] = encapsulate_streaming_html(
      [html_src], compress=compress, stealable=False
  )
  return converted.html_src



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reflection-based handlers for functions, closures, and classes.

These handlers provide:
- deep-linking to jump to the place a particular object was defined,
- visualization of closure variables for closures
"""

import functools
import inspect
import re
from typing import Any

from treescope import renderers
from treescope import rendering_parts


@functools.cache
def _get_filepath_and_lineno(value) -> tuple[str, int] | tuple[None, None]:
  """Retrieves a filepath and line number for an inspectable value."""
  try:
    filepath = inspect.getsourcefile(value)
    if filepath is None:
      return None, None
    # This also returns the source lines, but we don't need those.
    _, lineno = inspect.findsource(value)
    # Add 1, since findsource is zero-indexed.
    return filepath, lineno + 1
  except (TypeError, OSError):
    return None, None


def format_source_location(
    filepath: str, lineno: int
) -> rendering_parts.RenderableTreePart:
  """Formats a reference to a given filepath and line number."""

  # Try to match it as an IPython file
  ipython_output_path = re.fullmatch(
      r"<ipython-input-(?P<cell_number>\d+)-.*>", filepath
  )
  if ipython_output_path:
    cell_number = ipython_output_path.group("cell_number")
    return rendering_parts.text(f"line {lineno} of output cell {cell_number}")

  return rendering_parts.text(f"line {lineno} of {filepath}")


def handle_code_objects_with_reflection(
    node: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
    show_closure_vars: bool = False,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders code objects using reflection and closure inspection."""
  if inspect.isclass(node):
    # Render class.
    closure_vars = None
  elif inspect.isfunction(node):
    # Render function.
    if show_closure_vars:
      closure_vars = inspect.getclosurevars(node).nonlocals
    else:
      closure_vars = None
  else:
    # Not a supported object.
    return NotImplemented

  annotations = []
  filepath, lineno = _get_filepath_and_lineno(node)
  if filepath is not None:
    annotations.append(
        rendering_parts.comment_color(
            rendering_parts.siblings(
                rendering_parts.text("  # Defined at "),
                format_source_location(filepath, lineno),
            )
        )
    )

  if closure_vars:
    boxed_closure_var_rendering = rendering_parts.in_outlined_box(
        rendering_parts.on_separate_lines([
            rendering_parts.comment_color(
                rendering_parts.text("# Closure variables:")
            ),
            subtree_renderer(closure_vars),
        ])
    )
    return rendering_parts.siblings_with_annotations(
        rendering_parts.build_custom_foldable_tree_node(
            label=rendering_parts.abbreviation_color(
                rendering_parts.text(repr(node))
            ),
            contents=rendering_parts.fold_condition(
                expanded=rendering_parts.indented_children(
                    [boxed_closure_var_rendering]
                )
            ),
            path=path,
        ),
        extra_annotations=annotations,
    )

  else:
    return rendering_parts.siblings_with_annotations(
        rendering_parts.build_one_line_tree_node(
            line=rendering_parts.abbreviation_color(
                rendering_parts.text(repr(node))
            ),
            path=path,
        ),
        extra_annotations=annotations,
    )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the treescope renderer and handlers."""

import ast
import collections
import dataclasses
import functools
import re
import textwrap
import types
from typing import Any, Callable
import warnings

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
import numpy as np
import torch
import treescope
from treescope import handlers
from treescope import layout_algorithms
from treescope import lowering
from treescope import rendering_parts
from treescope.external import jax_support
from tests.fixtures import treescope_examples_fixture as fixture_lib


@dataclasses.dataclass
class CustomReprHTMLObject:
  repr_html: str

  def _repr_html_(self):
    return self.repr_html


class TreescopeRendererTest(parameterized.TestCase):

  def test_renderer_interface(self):
    renderer = treescope.active_renderer.get()

    rendering = renderer.to_text({"key": "value"})
    self.assertEqual(rendering, "{'key': 'value'}")

    rendering = renderer.to_html({"key": "value"})
    self.assertIsInstance(rendering, str)

    rendering = renderer.to_foldable_representation({"key": "value"})
    self.assertIsInstance(
        rendering, rendering_parts.RenderableAndLineAnnotations
    )

  def test_high_level_interface(self):
    rendering = treescope.render_to_text({"key": "value"})
    self.assertEqual(rendering, "{'key': 'value'}")

    rendering = treescope.render_to_html({"key": "value"})
    self.assertIsInstance(rendering, str)

  def test_error_recovery(self):
    def handler_that_crashes(node, path, subtree_renderer):
      del path, subtree_renderer
      if node == "trigger handler error":
        raise RuntimeError("handler error!")
      return NotImplemented

    def hook_that_crashes(node, path, node_renderer):
      del path, node_renderer
      if node == "trigger hook error":
        raise RuntimeError("hook error!")
      return NotImplemented

    renderer = treescope.active_renderer.get().extended_with(
        handlers=[handler_that_crashes], wrapper_hooks=[hook_that_crashes]
    )

    rendering = renderer.to_foldable_representation([1, 2, 3, "foo", 4])
    layout_algorithms.expand_to_depth(rendering.renderable, 1)
    self.assertEqual(
        lowering.render_to_text_as_root(
            rendering_parts.build_full_line_with_annotations(rendering)
        ),
        "[\n  1,\n  2,\n  3,\n  'foo',\n  4,\n]",
    )

    with self.assertRaisesWithLiteralMatch(RuntimeError, "handler error!"):
      _ = renderer.to_foldable_representation(
          [1, 2, 3, "trigger handler error", 4]
      )

    with self.assertRaisesWithLiteralMatch(RuntimeError, "hook error!"):
      _ = renderer.to_foldable_representation(
          [1, 2, 3, "trigger hook error", 4]
      )

    with warnings.catch_warnings(record=True) as recorded:
      rendering = renderer.to_foldable_representation(
          [1, 2, 3, "trigger handler error", "trigger hook error", 4],
          ignore_exceptions=True,
      )
    layout_algorithms.expand_to_depth(rendering.renderable, 1)
    self.assertEqual(
        lowering.render_to_text_as_root(
            rendering_parts.build_full_line_with_annotations(rendering)
        ),
        "[\n  1,\n  2,\n  3,\n  'trigger handler error',\n  'trigger hook"
        " error',\n  4,\n]",
    )
    self.assertLen(recorded, 2)
    self.assertEqual(type(recorded[0]), warnings.WarningMessage)
    self.assertContainsInOrder(
        [
            "Ignoring error while formatting value",
            "with",
            "<locals>.handler_that_crashes",
        ],
        str(recorded[0].message),
    )
    self.assertEqual(type(recorded[1]), warnings.WarningMessage)
    self.assertContainsInOrder(
        [
            "Ignoring error inside wrapper hook",
            "<locals>.hook_that_crashes",
        ],
        str(recorded[1].message),
    )

  @parameterized.named_parameters(
      dict(
          testcase_name="literals",
          target=[False, True, None, Ellipsis, NotImplemented],
          expected_collapsed="[False, True, None, Ellipsis, NotImplemented]",
      ),
      dict(
          testcase_name="numbers",
          target=[1234, 1234.0],
          expected_collapsed="[1234, 1234.0]",
      ),
      dict(
          testcase_name="strings",
          target=["some string", b"some bytes"],
          expected_collapsed="['some string', b'some bytes']",
      ),
      dict(
          testcase_name="multiline_string",
          target="some string\n    with \n newlines in it",
          expected_collapsed="'some string\\n    with \\n newlines in it'",
          expected_expanded=(
              "  'some string\\n'\n  '    with \\n'\n  ' newlines in it'"
          ),
      ),
      dict(
          testcase_name="enums",
          target=[
              fixture_lib.MyTestEnum.FOO,
              fixture_lib.MyTestEnum.BAR,
          ],
          expected_collapsed="[MyTestEnum.FOO, MyTestEnum.BAR]",
          expected_expanded=textwrap.dedent("""\
              [
                MyTestEnum.FOO,  # value: 1
                MyTestEnum.BAR,  # value: 2
              ]"""),
          expected_roundtrip=textwrap.dedent("""\
              [
                tests.fixtures.treescope_examples_fixture.MyTestEnum.FOO,  # value: 1
                tests.fixtures.treescope_examples_fixture.MyTestEnum.BAR,  # value: 2
              ]"""),
      ),
      dict(
          testcase_name="empty_dict",
          target={},
          expected_collapsed="{}",
      ),
      dict(
          testcase_name="dict_with_contents",
          target={"a": "b", (1, 2, 3, 4): (5, 6, 7, 8), (): (9, 10)},
          expected_collapsed=(
              "{'a': 'b', (1, 2, 3, 4): (5, 6, 7, 8), (): (9, 10)}"
          ),
          expected_expanded=textwrap.dedent("""\
              {
                'a': 'b',
                (1, 2, 3, 4):
                  (5, 6, 7, 8),
                (): (9, 10),
              }"""),
      ),
      dict(
          testcase_name="dict_subclass",
          target=collections.OrderedDict({"a": "b", (): (9, 10)}),
          expected_collapsed="OrderedDict({'a': 'b', (): (9, 10)})",
          expected_expanded=textwrap.dedent("""\
              OrderedDict({
                'a': 'b',
                (): (9, 10),
              })"""),
          expected_roundtrip=textwrap.dedent("""\
              collections.OrderedDict({
                'a': 'b',
                (): (9, 10),
              })"""),
      ),
      dict(
          testcase_name="tuple_empty",
          target=(),
          expected_collapsed="()",
      ),
      dict(
          testcase_name="tuple_singleton",
          target=(1,),
          expected_collapsed="(1,)",
          expected_expanded=textwrap.dedent("""\
              (
                1,
              )"""),
      ),
      dict(
          testcase_name="tuple_multiple",
          target=(1, 2, 3),
          expected_collapsed="(1, 2, 3)",
          expected_expanded=textwrap.dedent("""\
              (
                1,
                2,
                3,
              )"""),
      ),
      dict(
          testcase_name="list_empty",
          target=[],
          expected_collapsed="[]",
      ),
      dict(
          testcase_name="list_singleton",
          target=[1],
          expected_collapsed="[1]",
          expected_expanded=textwrap.dedent("""\
              [
                1,
              ]"""),
      ),
      dict(
          testcase_name="list_multiple",
          target=[1, 2, 3],
          expected_collapsed="[1, 2, 3]",
          expected_expanded=textwrap.dedent("""\
              [
                1,
                2,
                3,
              ]"""),
      ),
      dict(
          testcase_name="sets_empty",
          target=[set(), frozenset()],
          expected_collapsed="[set(), frozenset({})]",
          expected_expanded=textwrap.dedent("""\
              [
                set(),
                frozenset({}),
              ]"""),
      ),
      dict(
          testcase_name="set_with_items",
          target={1, 2, 3},
          expected_collapsed="{1, 2, 3}",
          expected_expanded=textwrap.dedent("""\
              {
                1,
                2,
                3,
              }"""),
      ),
      dict(
          testcase_name="frozenset_with_items",
          target=frozenset({1, 2, 3}),
          expected_collapsed="frozenset({1, 2, 3})",
          expected_expanded=textwrap.dedent("""\
              frozenset({
                1,
                2,
                3,
              })"""),
      ),
      dict(
          testcase_name="simplenamespace",
          target=types.SimpleNamespace(foo="bar", baz="qux"),
          expected_collapsed="SimpleNamespace(foo='bar', baz='qux')",
          expected_expanded=textwrap.dedent("""\
              SimpleNamespace(
                foo='bar',
                baz='qux',
              )"""),
      ),
      dict(
          testcase_name="nametuple",
          target=fixture_lib.SomeNamedtupleClass(foo="baz", bar="qux"),
          expected_collapsed="SomeNamedtupleClass(foo='baz', bar='qux')",
          expected_expanded=textwrap.dedent("""\
              SomeNamedtupleClass(
                foo='baz',
                bar='qux',
              )"""),
          expected_roundtrip=textwrap.dedent("""\
              tests.fixtures.treescope_examples_fixture.SomeNamedtupleClass(
                foo='baz',
                bar='qux',
              )"""),
      ),
      dict(
          testcase_name="dataclass",
          target=fixture_lib.DataclassWithTwoChildren(foo="baz", bar="qux"),
          expected_collapsed="DataclassWithTwoChildren(foo='baz', bar='qux')",
          expected_expanded=textwrap.dedent("""\
              DataclassWithTwoChildren(
                foo='baz',
                bar='qux',
              )"""),
          expected_roundtrip=textwrap.dedent("""\
              tests.fixtures.treescope_examples_fixture.DataclassWithTwoChildren(
                foo='baz',
                bar='qux',
              )"""),
      ),
      dict(
          testcase_name="dataclass_empty",
          target=fixture_lib.EmptyDataclass(),
          expected_collapsed="EmptyDataclass()",
          expected_expanded="EmptyDataclass()",
          expected_roundtrip=(
              "tests.fixtures.treescope_examples_fixture.EmptyDataclass()"
          ),
      ),
      dict(
          testcase_name="ndarray_small",
          target=np.array([1, 2, 4, 8, 16]),
          expected_collapsed="np.array([ 1,  2,  4,  8, 16])",
          expected_expanded="np.array([ 1,  2,  4,  8, 16])",
      ),
      dict(
          testcase_name="ndarray_large",
          target=np.arange(3 * 7).reshape((3, 7)),
          expected_collapsed=(
              "<np.ndarray int64(3, 7) [≥0, ≤20] zero:1 nonzero:20>"
          ),
          expected_expanded=textwrap.dedent("""\
              # np.ndarray int64(3, 7) [≥0, ≤20] zero:1 nonzero:20
                array([[ 0,  1,  2,  3,  4,  5,  6],
                       [ 7,  8,  9, 10, 11, 12, 13],
                       [14, 15, 16, 17, 18, 19, 20]])"""),
      ),
      dict(
          testcase_name="jax_array_large",
          target_builder=lambda: jnp.arange(3 * 7).reshape((3, 7)),
          expected_collapsed=(
              "<jax.Array int32(3, 7) [≥0, ≤20] zero:1 nonzero:20>"
          ),
          expected_expanded=textwrap.dedent("""\
              # jax.Array int32(3, 7) [≥0, ≤20] zero:1 nonzero:20
                Array([[ 0,  1,  2,  3,  4,  5,  6],
                       [ 7,  8,  9, 10, 11, 12, 13],
                       [14, 15, 16, 17, 18, 19, 20]], dtype=int32)"""),
      ),
      dict(
          testcase_name="pytorch_tensor_small",
          target_builder=lambda: torch.tensor(np.array([1, 2, 4, 8, 16])),
          expected_collapsed="torch.tensor([ 1,  2,  4,  8, 16])",
          expected_expanded="torch.tensor([ 1,  2,  4,  8, 16])",
      ),
      dict(
          testcase_name="pytorch_tensor_large",
          target_builder=lambda: torch.tensor(np.arange(3 * 7).reshape((3, 7))),
          expected_collapsed=(
              "<torch.Tensor int64(3, 7) [≥0, ≤20] zero:1 nonzero:20>"
          ),
          expected_expanded=textwrap.dedent("""\
              # torch.Tensor int64(3, 7) [≥0, ≤20] zero:1 nonzero:20
                tensor([[ 0,  1,  2,  3,  4,  5,  6],
                        [ 7,  8,  9, 10, 11, 12, 13],
                        [14, 15, 16, 17, 18, 19, 20]])"""),
      ),
      dict(
          testcase_name="well_known_function",
          target=treescope.render_to_text,
          expected_collapsed="render_to_text",
          expected_roundtrip_collapsed="treescope.render_to_text",
      ),
      dict(
          testcase_name="well_known_type",
          target=treescope.IPythonVisualization,
          expected_collapsed="IPythonVisualization",
          expected_roundtrip_collapsed="treescope.IPythonVisualization",
      ),
      dict(
          testcase_name="ast_nodes",
          target=ast.parse("print(1, 2)").body[0],
          expand_depth=3,
          expected_expanded=textwrap.dedent("""\
              Expr(
                value=Call(
                  func=Name(
                    id='print',
                    ctx=Load(),
                  ),
                  args=[
                    Constant(value=1, kind=None),
                    Constant(value=2, kind=None),
                  ],
                  keywords=[],
                ),
              )"""),
      ),
      dict(
          testcase_name="custom_handler",
          target=[fixture_lib.ObjectWithCustomHandler()],
          expected_collapsed=(
              "[<ObjectWithCustomHandler custom rendering! Path: '[0]'>]"
          ),
      ),
      dict(
          testcase_name="custom_handler_that_throws",
          target=[fixture_lib.ObjectWithCustomHandlerThatThrows()],
          ignore_exceptions=True,
          expected_collapsed=(
              "[<Fallback repr for ObjectWithCustomHandlerThatThrows>]"
          ),
      ),
      dict(
          testcase_name="dtype_standard",
          target=np.dtype(np.float32),
          expected_collapsed="dtype('float32')",
          expected_roundtrip_collapsed="np.dtype('float32')",
      ),
      dict(
          testcase_name="dtype_extended",
          target=np.dtype(jnp.bfloat16),
          expected_collapsed="dtype('bfloat16')",
          expected_roundtrip_collapsed="np.dtype('bfloat16')",
      ),
      dict(
          testcase_name="jax_precision",
          target=[jax.lax.Precision.HIGHEST],
          expected_collapsed="[Precision.HIGHEST]",
          expected_expanded=textwrap.dedent("""\
              [
                Precision.HIGHEST,  # value: 2
              ]"""),
          expected_roundtrip_collapsed="[jax.lax.Precision.HIGHEST]",
      ),
      dict(
          testcase_name="pytorch_module",
          target_builder=fixture_lib.SomePyTorchModule.build,
          expected_collapsed=(
              "SomePyTorchModule(attr_one=123, attr_two='abc', training=True,"
              " foo=<torch.nn.Parameter float32(5,) ≈1.0 ±0.0 [≥1.0, ≤1.0]"
              " nonzero:5>, bar=torch.tensor([0., 0., 0., 0., 0.]),"
              " linear=Linear(in_features=10, out_features=10, training=True,"
              " weight=<torch.nn.Parameter float32(10, 10) ≈-0.008 ±0.028"
              " [≥-0.32, ≤0.3] nonzero:100>, bias=<torch.nn.Parameter"
              " float32(10,) ≈0.008 ±0.04 [≥-0.31, ≤0.31] nonzero:10>, ),"
              " mod_list=ModuleList(training=True, (0):"
              " LayerNorm(normalized_shape=(10,), eps=1e-05,"
              " elementwise_affine=True, training=True,"
              " weight=<torch.nn.Parameter float32(10,) ≈1.0 ±0.0 [≥1.0, ≤1.0]"
              " nonzero:10>, bias=<torch.nn.Parameter float32(10,) ≈0.0 ±0.0"
              " [≥0.0, ≤0.0] zero:10>, ), (1): SiLU(inplace=False,"
              " training=True, ), ), )"
          ),
          expected_expanded=textwrap.dedent("""\
              SomePyTorchModule(
                attr_one=123, attr_two='abc', training=True,
                # Parameters:
                foo=<torch.nn.Parameter float32(5,) ≈1.0 ±0.0 [≥1.0, ≤1.0] nonzero:5>,
                # Buffers:
                bar=torch.tensor([0., 0., 0., 0., 0.]),
                # Child modules:
                linear=Linear(in_features=10, out_features=10, training=True, weight=<torch.nn.Parameter float32(10, 10) ≈-0.008 ±0.028 [≥-0.32, ≤0.3] nonzero:100>, bias=<torch.nn.Parameter float32(10,) ≈0.008 ±0.04 [≥-0.31, ≤0.31] nonzero:10>, ),
                mod_list=ModuleList(training=True, (0): LayerNorm(normalized_shape=(10,), eps=1e-05, elementwise_affine=True, training=True, weight=<torch.nn.Parameter float32(10,) ≈1.0 ±0.0 [≥1.0, ≤1.0] nonzero:10>, bias=<torch.nn.Parameter float32(10,) ≈0.0 ±0.0 [≥0.0, ≤0.0] zero:10>, ), (1): SiLU(inplace=False, training=True, ), ),
              )"""),
          expected_roundtrip=textwrap.dedent("""\
              <tests.fixtures.treescope_examples_fixture.SomePyTorchModule(
                attr_one=123, attr_two='abc', training=True,
                # Parameters:
                foo=<torch.nn.Parameter float32(5,) ≈1.0 ±0.0 [≥1.0, ≤1.0] nonzero:5>,
                # Buffers:
                bar=torch.tensor([0., 0., 0., 0., 0.]),
                # Child modules:
                linear=<torch.nn.modules.linear.Linear(in_features=10, out_features=10, training=True, weight=<torch.nn.Parameter float32(10, 10) ≈-0.008 ±0.028 [≥-0.32, ≤0.3] nonzero:100>, bias=<torch.nn.Parameter float32(10,) ≈0.008 ±0.04 [≥-0.31, ≤0.31] nonzero:10>, )>,
                mod_list=<torch.nn.modules.container.ModuleList(training=True, (0): <torch.nn.modules.normalization.LayerNorm(normalized_shape=(10,), eps=1e-05, elementwise_affine=True, training=True, weight=<torch.nn.Parameter float32(10,) ≈1.0 ±0.0 [≥1.0, ≤1.0] nonzero:10>, bias=<torch.nn.Parameter float32(10,) ≈0.0 ±0.0 [≥0.0, ≤0.0] zero:10>, )>, (1): <torch.nn.modules.activation.SiLU(inplace=False, training=True, )>, )>,
              )>"""),
      ),
      dict(
          testcase_name="pytorch_module_expanded",
          target_builder=fixture_lib.SomePyTorchModule.build,
          expand_depth=2,
          expected_expanded=textwrap.dedent("""\
              SomePyTorchModule(
                # Attributes:
                attr_one=123,
                attr_two='abc',
                training=True,
                # Parameters:
                foo=# torch.nn.Parameter float32(5,) ≈1.0 ±0.0 [≥1.0, ≤1.0] nonzero:5
                  Parameter containing:
                  tensor([1., 1., 1., 1., 1.], requires_grad=True)
                ,
                # Buffers:
                bar=torch.tensor([0., 0., 0., 0., 0.]),
                # Child modules:
                linear=Linear(
                  in_features=10, out_features=10, training=True,
                  # Parameters:
                  weight=<torch.nn.Parameter float32(10, 10) ≈-0.008 ±0.028 [≥-0.32, ≤0.3] nonzero:100>,
                  bias=<torch.nn.Parameter float32(10,) ≈0.008 ±0.04 [≥-0.31, ≤0.31] nonzero:10>,
                ),
                mod_list=ModuleList(
                  training=True,
                  # Child modules:
                  (0): LayerNorm(normalized_shape=(10,), eps=1e-05, elementwise_affine=True, training=True, weight=<torch.nn.Parameter float32(10,) ≈1.0 ±0.0 [≥1.0, ≤1.0] nonzero:10>, bias=<torch.nn.Parameter float32(10,) ≈0.0 ±0.0 [≥0.0, ≤0.0] zero:10>, ),
                  (1): SiLU(inplace=False, training=True, ),
                ),
              )"""),
      ),
  )
  def test_object_rendering(
      self,
      *,
      target: Any = None,
      target_builder: Callable[[], Any] | None = None,
      expected_collapsed: str | None = None,
      expected_expanded: str | None = None,
      expected_roundtrip: str | None = None,
      expected_roundtrip_collapsed: str | None = None,
      expand_depth: int = 1,
      ignore_exceptions: bool = False,
  ):
    if target_builder is not None:
      assert target is None
      target = target_builder()

    renderer = treescope.active_renderer.get()
    # Render it to IR.
    with warnings.catch_warnings():
      if ignore_exceptions:
        # Also ignore warnings due to ignoring exceptions to avoid cluttering
        # logs.
        warnings.simplefilter("ignore")
      rendering = rendering_parts.build_full_line_with_annotations(
          renderer.to_foldable_representation(
              target, ignore_exceptions=ignore_exceptions
          )
      )

    # Collapse all foldables.
    layout_algorithms.expand_to_depth(rendering, 0)

    if expected_collapsed is not None:
      with self.subTest("collapsed"):
        self.assertEqual(
            lowering.render_to_text_as_root(rendering),
            expected_collapsed,
        )

    if expected_roundtrip_collapsed is not None:
      with self.subTest("roundtrip_collapsed"):
        self.assertEqual(
            lowering.render_to_text_as_root(rendering, roundtrip=True),
            expected_roundtrip_collapsed,
        )

    layout_algorithms.expand_to_depth(rendering, expand_depth)

    if expected_expanded is not None:
      with self.subTest("expanded"):
        self.assertEqual(
            lowering.render_to_text_as_root(rendering),
            expected_expanded,
        )

    if expected_roundtrip is not None:
      with self.subTest("roundtrip"):
        self.assertEqual(
            lowering.render_to_text_as_root(rendering, roundtrip=True),
            expected_roundtrip,
        )

    # Render to HTML; make sure it doesn't raise any errors.
    with self.subTest("html_no_errors"):
      _ = lowering.render_to_html_as_root(rendering)

  def test_closure_rendering(self):
    def outer_fn(x):
      def inner_fn(y):
        return x + y

      return inner_fn

    closure = outer_fn(100)

    renderer = treescope.active_renderer.get()
    # Enable closure rendering (currently disabled by default)
    renderer = renderer.extended_with(
        handlers=[
            functools.partial(
                handlers.handle_code_objects_with_reflection,
                show_closure_vars=True,
            )
        ]
    )
    # Render it to IR.
    rendering = rendering_parts.build_full_line_with_annotations(
        renderer.to_foldable_representation(closure)
    )

    layout_algorithms.expand_to_depth(rendering, 1)

    self.assertContainsInOrder(
        [
            "<function",
            "test_closure_rendering.<locals>.outer_fn.<locals>.inner_fn at 0x",
            ">",
            "# Closure variables:",
            "{'x': 100}",
            "# Defined at line ",
            " of ",
            "tests/renderer_test.py",
        ],
        lowering.render_to_text_as_root(rendering),
    )

  def test_render_jax_array_within_jitted_function(self):
    old = jax_support.SUMMARIZE_USING_NUMPY_THRESHOLD
    jax_support.SUMMARIZE_USING_NUMPY_THRESHOLD = 0
    renderer = treescope.active_renderer.get()
    x = jnp.arange(10)

    # Verify that we don't fail when called inside a jitted function.
    @jax.jit
    def go(s):
      nonlocal renderer, x
      adapter = jax_support.JAXArrayAdapter()
      self.assertNotEmpty(adapter.get_array_summary(x, False))
      return jax.numpy.sum(s)

    go(jnp.arange(3))
    jax_support.SUMMARIZE_USING_NUMPY_THRESHOLD = old

  def test_fallback_repr_pytree_node(self):
    target = [fixture_lib.UnknownPytreeNode(1234, 5678)]
    renderer = treescope.active_renderer.get()
    rendering = rendering_parts.build_full_line_with_annotations(
        renderer.to_foldable_representation(target)
    )
    layout_algorithms.expand_to_depth(rendering, 0)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        "[<custom repr for UnknownPytreeNode: x=1234, y=5678>]",
    )

    layout_algorithms.expand_to_depth(rendering, 2)
    rendered_text = lowering.render_to_text_as_root(rendering)
    self.assertEqual(
        "\n".join(
            line.rstrip() for line in rendered_text.splitlines(keepends=True)
        ),
        textwrap.dedent(f"""\
            [
              <custom repr for UnknownPytreeNode: x=1234, y=5678>
                #╭┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╮
                # PyTree children:
                  FrozenDataclassKey(name='x'): 1234,
                  'string_key': 5678,
                #╰┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╯
              ,  # {object.__repr__(target[0])}
            ]"""),
    )

  def test_fallback_repr_one_line(self):
    target = [fixture_lib.UnknownObjectWithOneLineRepr()]
    renderer = treescope.active_renderer.get()
    rendering = rendering_parts.build_full_line_with_annotations(
        renderer.to_foldable_representation(target)
    )
    layout_algorithms.expand_to_depth(rendering, 0)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        "[<custom repr for UnknownObjectWithOneLineRepr>]",
    )
    layout_algorithms.expand_to_depth(rendering, 2)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent(f"""\
            [
              <custom repr for UnknownObjectWithOneLineRepr>,  # {object.__repr__(target[0])}
            ]"""),
    )

  def test_fallback_repr_after_error(self):
    target = [fixture_lib.ObjectWithCustomHandlerThatThrows()]
    renderer = treescope.active_renderer.get()

    with self.assertRaisesWithLiteralMatch(
        RuntimeError, "Simulated treescope_repr failure!"
    ):
      renderer.to_foldable_representation(target)

    with self.assertWarnsRegex(
        UserWarning,
        "(.|\n)*".join([
            re.escape("Ignoring error while formatting value of type"),
            re.escape("ObjectWithCustomHandlerThatThrows"),
            re.escape(
                'raise RuntimeError("Simulated treescope_repr failure!")'
            ),
        ]),
    ):
      rendering = rendering_parts.build_full_line_with_annotations(
          renderer.to_foldable_representation(target, ignore_exceptions=True)
      )

    layout_algorithms.expand_to_depth(rendering, 0)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        "[<Fallback repr for ObjectWithCustomHandlerThatThrows>]",
    )
    layout_algorithms.expand_to_depth(rendering, 2)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent(f"""\
            [
              <Fallback repr for ObjectWithCustomHandlerThatThrows>,  # {object.__repr__(target[0])}
            ]"""),
    )

  def test_ignore_exceptions_in_deferred(self):
    target = [fixture_lib.ObjectWithCustomHandlerThatThrowsDeferred()]
    renderer = treescope.active_renderer.get()

    with self.assertRaisesWithLiteralMatch(
        RuntimeError, "Simulated deferred treescope_repr failure!"
    ):
      renderer.to_foldable_representation(target)

    with lowering.collecting_deferred_renderings() as deferreds:
      foldable_ir = rendering_parts.build_full_line_with_annotations(
          renderer.to_foldable_representation(target)
      )

    # It's difficult to test the IPython wrapper so we instead test the internal
    # helper function that produces the streaming HTML output.
    html_parts = lowering._render_to_html_as_root_streaming(
        root_node=foldable_ir,
        roundtrip=False,
        deferreds=deferreds,
        ignore_exceptions=True,
    )
    self.assertContainsInOrder(
        [
            "[",
            "&lt;RuntimeError during deferred rendering",
            "Traceback",
            "in _internal_main_thunk",
            "raise RuntimeError",
            "RuntimeError: Simulated deferred treescope_repr failure!",
            "&gt;",
            "]",
        ],
        "".join(html_parts),
    )

  def test_fallback_repr_multiline_idiomatic(self):
    target = [fixture_lib.UnknownObjectWithMultiLineRepr()]
    renderer = treescope.active_renderer.get()
    rendering = rendering_parts.build_full_line_with_annotations(
        renderer.to_foldable_representation(target)
    )
    layout_algorithms.expand_to_depth(rendering, 0)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        "[<custom repr↩  for↩  UnknownObjectWithMultiLineRepr↩>]",
    )
    layout_algorithms.expand_to_depth(rendering, 2)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent(f"""\
            [
              <custom repr
                for
                UnknownObjectWithMultiLineRepr
              >,  # {object.__repr__(target[0])}
            ]"""),
    )

  def test_fallback_repr_multiline_unidiomatic(self):
    target = [fixture_lib.UnknownObjectWithBadMultiLineRepr()]
    renderer = treescope.active_renderer.get()
    rendering = rendering_parts.build_full_line_with_annotations(
        renderer.to_foldable_representation(target)
    )
    layout_algorithms.expand_to_depth(rendering, 0)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        f"[{object.__repr__(target[0])}]",
    )
    layout_algorithms.expand_to_depth(rendering, 2)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent(f"""\
            [
              # {object.__repr__(target[0])}
                Non-idiomatic
                multiline
                object
              ,
            ]"""),
    )

  def test_fallback_repr_basic(self):
    target = [fixture_lib.UnknownObjectWithBuiltinRepr()]
    renderer = treescope.active_renderer.get()
    rendering = rendering_parts.build_full_line_with_annotations(
        renderer.to_foldable_representation(target)
    )
    layout_algorithms.expand_to_depth(rendering, 0)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        f"[{repr(target[0])}]",
    )
    layout_algorithms.expand_to_depth(rendering, 2)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent(f"""\
            [
              {repr(target[0])},
            ]"""),
    )

  def test_failsafe_for_throw_in_repr(self):
    target = [fixture_lib.ObjectWithReprThatThrows()]
    renderer = treescope.active_renderer.get()

    with self.assertRaisesWithLiteralMatch(
        RuntimeError, "Simulated repr failure!"
    ):
      renderer.to_foldable_representation(target)

    with self.assertWarnsRegex(
        UserWarning,
        "(.|\n)*".join([
            re.escape("Ignoring error while formatting value of type"),
            re.escape("ObjectWithReprThatThrows"),
            re.escape('raise RuntimeError("Simulated repr failure!")'),
        ]),
    ):
      rendering = rendering_parts.build_full_line_with_annotations(
          renderer.to_foldable_representation(target, ignore_exceptions=True)
      )

    layout_algorithms.expand_to_depth(rendering, 0)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        f"[{object.__repr__(target[0])}]",
    )
    layout_algorithms.expand_to_depth(rendering, 2)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent(f"""\
            [
              {object.__repr__(target[0])},  # Error occurred while formatting this object.
            ]"""),
    )

  def test_shared_values(self):
    shared = ["bar"]
    target = [shared, shared, {"foo": shared}]
    renderer = treescope.active_renderer.get()
    rendering = rendering_parts.build_full_line_with_annotations(
        renderer.to_foldable_representation(target)
    )
    layout_algorithms.expand_to_depth(rendering, 3)
    rendered_text = lowering.render_to_text_as_root(rendering)
    # Rendering may contain trailing whitespace; remove it before checking the
    # value, since it's not important.
    self.assertEqual(
        "\n".join(
            line.rstrip() for line in rendered_text.splitlines(keepends=True)
        ),
        textwrap.dedent(f"""\
            [
              [
                'bar',
              ], # Repeated python obj at 0x{id(shared):x}
              [
                'bar',
              ], # Repeated python obj at 0x{id(shared):x}
              {{
                'foo': [
                  'bar',
                ], # Repeated python obj at 0x{id(shared):x}
              }},
            ]"""),
    )

  def test_autovisualizer(self):
    target = [1, 2, "foo", 3, 4, 5, [6, 7]]

    def autovisualizer_for_test(node, path):
      if isinstance(node, str):
        return treescope.VisualizationFromTreescopePart(
            rendering_parts.RenderableAndLineAnnotations(
                rendering_parts.text("(visualization for foo goes here)"),
                rendering_parts.text(" # annotation for vis for foo"),
            ),
        )
      elif path == "[4]":
        return treescope.IPythonVisualization(
            CustomReprHTMLObject("(html rendering)"),
            replace=True,
        )
      elif path == "[5]":
        return treescope.IPythonVisualization(
            CustomReprHTMLObject("(html rendering)"),
            replace=False,
        )
      elif path == "[6]":
        return treescope.ChildAutovisualizer(inner_autovisualizer)

    def inner_autovisualizer(node, path):
      del path
      if node == 6:
        return treescope.VisualizationFromTreescopePart(
            rendering_parts.RenderableAndLineAnnotations(
                rendering_parts.text("(child visualization of 6 goes here)"),
                rendering_parts.text(" # annotation for vis for 6"),
            ),
        )

    with treescope.active_autovisualizer.set_scoped(autovisualizer_for_test):
      renderer = treescope.active_renderer.get()
      rendering = rendering_parts.build_full_line_with_annotations(
          renderer.to_foldable_representation(target)
      )
      layout_algorithms.expand_to_depth(rendering, 3)
      rendered_text = lowering.render_to_text_as_root(rendering)
      rendered_text_as_roundtrip = lowering.render_to_text_as_root(
          rendering, roundtrip=True
      )

    self.assertEqual(
        rendered_text,
        textwrap.dedent("""\
            [
              1,
              2,
              (visualization for foo goes here), # annotation for vis for foo
              3,
              <Visualization of int:
                <rich HTML visualization>
              >,
              5
              #╭┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╮
              <rich HTML visualization>
              #╰┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╯
              ,
              [
                (child visualization of 6 goes here), # annotation for vis for 6
                7,
              ],
            ]"""),
    )

    self.assertEqual(
        rendered_text_as_roundtrip,
        textwrap.dedent("""\
            [
              1,
              2,
              'foo',  # Visualization hidden in roundtrip mode
              3,
              4,  # Visualization hidden in roundtrip mode
              5
              #╭┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╮
              # <rich HTML visualization>
              #╰┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╯
              ,
              [
                6,  # Visualization hidden in roundtrip mode
                7,
              ],
            ]"""),
    )

  def test_balanced_layout(self):
    renderer = treescope.active_renderer.get()
    some_nested_object = fixture_lib.DataclassWithOneChild([
        ["foo"] * 4,
        ["12345678901234567890"] * 5,
        {"a": 1, "b": 2, "c": [{"bar": "baz"} for _ in range(5)]},
        [list(range(10)) for _ in range(6)],
    ])

    def render_and_expand(**kwargs):
      rendering = renderer.to_foldable_representation(
          some_nested_object
      ).renderable
      layout_algorithms.expand_for_balanced_layout(rendering, **kwargs)
      return lowering.render_to_text_as_root(rendering)

    with self.subTest("no_max_height"):
      self.assertEqual(
          render_and_expand(max_height=None, target_width=60),
          textwrap.dedent("""\
              DataclassWithOneChild(
                foo=[
                  ['foo', 'foo', 'foo', 'foo'],
                  [
                    '12345678901234567890',
                    '12345678901234567890',
                    '12345678901234567890',
                    '12345678901234567890',
                    '12345678901234567890',
                  ],
                  {
                    'a': 1,
                    'b': 2,
                    'c': [
                      {'bar': 'baz'},
                      {'bar': 'baz'},
                      {'bar': 'baz'},
                      {'bar': 'baz'},
                      {'bar': 'baz'},
                    ],
                  },
                  [
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                  ],
                ],
              )"""),
      )

    with self.subTest("medium_max_height"):
      self.assertEqual(
          render_and_expand(max_height=20, target_width=60),
          textwrap.dedent("""\
              DataclassWithOneChild(
                foo=[
                  ['foo', 'foo', 'foo', 'foo'],
                  [
                    '12345678901234567890',
                    '12345678901234567890',
                    '12345678901234567890',
                    '12345678901234567890',
                    '12345678901234567890',
                  ],
                  {
                    'a': 1,
                    'b': 2,
                    'c': [{'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}],
                  },
                  [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]],
                ],
              )"""),
      )

    with self.subTest("small_max_height"):
      self.assertEqual(
          render_and_expand(max_height=10, target_width=60),
          textwrap.dedent("""\
              DataclassWithOneChild(
                foo=[
                  ['foo', 'foo', 'foo', 'foo'],
                  ['12345678901234567890', '12345678901234567890', '12345678901234567890', '12345678901234567890', '12345678901234567890'],
                  {'a': 1, 'b': 2, 'c': [{'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}]},
                  [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]],
                ],
              )"""),
      )

    with self.subTest("long_target_width"):
      self.assertEqual(
          render_and_expand(max_height=None, target_width=150),
          textwrap.dedent("""\
              DataclassWithOneChild(
                foo=[
                  ['foo', 'foo', 'foo', 'foo'],
                  ['12345678901234567890', '12345678901234567890', '12345678901234567890', '12345678901234567890', '12345678901234567890'],
                  {'a': 1, 'b': 2, 'c': [{'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}, {'bar': 'baz'}]},
                  [
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                  ],
                ],
              )"""),
      )

  def test_balanced_layout_after_manual_expansion(self):
    renderer = treescope.active_renderer.get()
    some_nested_object = [
        fixture_lib.DataclassWithOneChild(
            [["foo"] * 4, (["baz"] * 5, ["qux"] * 5)]
        )
    ]

    rendering = renderer.to_foldable_representation(
        some_nested_object
    ).renderable
    layout_algorithms.expand_for_balanced_layout(
        rendering,
        max_height=3,
        target_width=40,
        recursive_expand_height_for_collapsed_nodes=10,
    )

    # Initially collapsed due to height constraint.
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent("""\
        [
          DataclassWithOneChild(foo=[['foo', 'foo', 'foo', 'foo'], (['baz', 'baz', 'baz', 'baz', 'baz'], ['qux', 'qux', 'qux', 'qux', 'qux'])]),
        ]"""),
    )

    # But expands multiple levels once we expand the collapsed node manually.
    # (In the browser, this would be done by clicking the expand marker.)
    target_foldable = (
        rendering.foldables_in_this_part()[0]
        .as_expanded_part()
        .foldables_in_this_part()[0]
    )
    self.assertEqual(
        target_foldable.get_expand_state(),
        rendering_parts.ExpandState.COLLAPSED,
    )
    target_foldable.set_expand_state(rendering_parts.ExpandState.EXPANDED)
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent("""\
            [
              DataclassWithOneChild(
                foo=[
                  ['foo', 'foo', 'foo', 'foo'],
                  (
                    ['baz', 'baz', 'baz', 'baz', 'baz'],
                    ['qux', 'qux', 'qux', 'qux', 'qux'],
                  ),
                ],
              ),
            ]"""),
    )

  def test_balanced_layout_relaxes_height_constraint_once(self):
    renderer = treescope.active_renderer.get()
    some_nested_object = [
        fixture_lib.DataclassWithOneChild(
            [fixture_lib.DataclassWithOneChild(["abcdefghik"] * 20)]
        )
    ]

    # With a relaxed only-child expansion constraint (the default), we still
    # expand the large list because it's the only nontrivial object.
    rendering = renderer.to_foldable_representation(
        some_nested_object
    ).renderable
    layout_algorithms.expand_for_balanced_layout(
        rendering, max_height=10, target_width=40
    )
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent("""\
            [
              DataclassWithOneChild(
                foo=[
                  DataclassWithOneChild(
                    foo=[
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                      'abcdefghik',
                    ],
                  ),
                ],
              ),
            ]"""),
    )

    # Without a relaxed only-child expansion constraint, it stays collapsed.
    rendering = renderer.to_foldable_representation(
        some_nested_object
    ).renderable
    layout_algorithms.expand_for_balanced_layout(
        rendering,
        max_height=10,
        target_width=40,
        relax_height_constraint_for_only_child=False,
    )
    self.assertEqual(
        lowering.render_to_text_as_root(rendering),
        textwrap.dedent("""\
            [
              DataclassWithOneChild(
                foo=[
                  DataclassWithOneChild(
                    foo=['abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik', 'abcdefghik'],
                  ),
                ],
              ),
            ]"""),
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Treescope rendering handlers, for use in renderer configurations.

These handlers are responsible for implementing each of the steps of the
default Treescope renderer, and can be used individually to build custom
renderer configurations.
"""

# pylint: disable=g-importing-member,g-multiple-import,unused-import

from treescope._internal.handlers.autovisualizer_hook import (
    use_autovisualizer_if_present,
)
from treescope._internal.handlers.basic_types_handler import (
    handle_basic_types,
)
from treescope._internal.handlers.canonical_alias_postprocessor import (
    replace_with_canonical_aliases,
)
from treescope._internal.handlers.custom_type_handlers import (
    handle_via_global_registry,
    handle_via_treescope_repr_method,
)
from treescope._internal.handlers.function_reflection_handlers import (
    handle_code_objects_with_reflection,
)
from treescope._internal.handlers.generic_pytree_handler import (
    handle_arbitrary_pytrees,
)
from treescope._internal.handlers.generic_repr_handler import (
    handle_anything_with_repr,
)
from treescope._internal.handlers.repr_html_postprocessor import (
    append_repr_html_when_present,
)
from treescope._internal.handlers.shared_value_postprocessor import (
    check_for_shared_values,
    setup_shared_value_context,
)

__all__ = [
    "use_autovisualizer_if_present",
    "handle_basic_types",
    "replace_with_canonical_aliases",
    "handle_via_global_registry",
    "handle_via_treescope_repr_method",
    "handle_code_objects_with_reflection",
    "handle_arbitrary_pytrees",
    "handle_anything_with_repr",
    "append_repr_html_when_present",
    "check_for_shared_values",
    "setup_shared_value_context",
]



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sphinx customizations for Treescope."""

from typing import Any, Mapping

from sphinx import application
from treescope._internal import docs_util

# pylint: disable=protected-access


def skip_member(app, what, name, obj, options, lines):  # pylint: disable=unused-argument
  """Configures autodoc to skip private implementation details."""
  if id(obj) in docs_util._SKIPPED_FOR_AUTODOC:
    return True


def setup(app: application.Sphinx) -> Mapping[str, Any]:
  app.connect("autodoc-skip-member", skip_member)
  return dict(version="0.1", parallel_read_safe=True)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers to build frequently-used configurations of rendered parts."""

from __future__ import annotations

from typing import Any, Sequence

from treescope import canonical_aliases
from treescope._internal.parts import basic_parts
from treescope._internal.parts import common_styles
from treescope._internal.parts import foldable_impl
from treescope._internal.parts import part_interface


CSSStyleRule = part_interface.CSSStyleRule
JavaScriptDefn = part_interface.JavaScriptDefn
HtmlContextForSetup = part_interface.HtmlContextForSetup
RenderableTreePart = part_interface.RenderableTreePart
ExpandState = part_interface.ExpandState
FoldableTreeNode = part_interface.FoldableTreeNode

RenderableAndLineAnnotations = basic_parts.RenderableAndLineAnnotations


def build_copy_button(path: str | None) -> RenderableTreePart:
  """Builds a copy-path button, if `path` is provided and not empty."""
  if not path:
    return basic_parts.EmptyPart()
  else:
    return foldable_impl.StringCopyButton(
        annotation="Copy path: ", copy_string=path
    )


def build_custom_foldable_tree_node(
    contents: RenderableTreePart,
    path: str | None = None,
    label: RenderableTreePart = basic_parts.EmptyPart(),
    expand_state: part_interface.ExpandState = (
        part_interface.ExpandState.WEAKLY_COLLAPSED
    ),
) -> RenderableAndLineAnnotations:
  """Builds a custom foldable tree node with path buttons.

  Args:
    contents: Contents of this foldable that should not open/close the custom
      foldable when clicked.
    path: Keypath to this node from the root. If provided, a copy-path button
      will be added at the end of the node.
    label: The beginning of the first line, which should allow opening/closing
      the custom foldable when clicked. Should not contain any other foldables.
    expand_state: Initial expand state for the foldable.

  Returns:
    A new renderable part, possibly with a copy button annotation, for use
    in part of a rendered treescope tree.
  """
  maybe_copy_button = build_copy_button(path)

  return RenderableAndLineAnnotations(
      renderable=foldable_impl.FoldableTreeNodeImpl(
          label=label, contents=contents, expand_state=expand_state
      ),
      annotations=maybe_copy_button,
  )


def build_one_line_tree_node(
    line: RenderableAndLineAnnotations | RenderableTreePart | str,
    path: str | None = None,
    background_color: str | None = None,
    background_pattern: str | None = None,
) -> RenderableAndLineAnnotations:
  """Builds a single-line tree node with path buttons.

  Args:
    line: Contents of the line.
    path: Keypath to this node from the root. If provided, copy-path buttons
      will be added.
    background_color: Optional background and border color for this node.
    background_pattern: Optional background pattern as a CSS "image". If
      provided, `background_color` must also be provided, and will be used as
      the border for the pattern.

  Returns:
    A new renderable part, possibly with a copy button annotation, for use
    in part of a rendered treescope tree.
  """
  maybe_copy_button = build_copy_button(path)

  if isinstance(line, RenderableAndLineAnnotations):
    line_primary = line.renderable
    annotations = basic_parts.siblings(maybe_copy_button, line.annotations)
  elif isinstance(line, str):
    line_primary = basic_parts.Text(line)
    annotations = maybe_copy_button
  else:
    line_primary = line
    annotations = maybe_copy_button

  if background_pattern is not None:
    if background_color is None:
      raise ValueError(
          "background_color must be provided if background_pattern is"
      )
    line_primary = common_styles.WithBlockPattern(
        common_styles.PatternedSingleLineSpanGroup(line_primary),
        color=background_color,
        pattern=background_pattern,
    )
  elif background_color is not None and background_color != "transparent":
    line_primary = common_styles.WithBlockColor(
        common_styles.ColoredSingleLineSpanGroup(line_primary),
        color=background_color,
    )

  return RenderableAndLineAnnotations(
      renderable=line_primary,
      annotations=annotations,
  )


def build_foldable_tree_node_from_children(
    prefix: RenderableTreePart | str,
    children: Sequence[RenderableAndLineAnnotations | RenderableTreePart | str],
    suffix: RenderableTreePart | str,
    comma_separated: bool = False,
    force_trailing_comma: bool = False,
    path: str | None = None,
    background_color: str | None = None,
    background_pattern: str | None = None,
    first_line_annotation: RenderableTreePart | None = None,
    expand_state: part_interface.ExpandState = (
        part_interface.ExpandState.WEAKLY_COLLAPSED
    ),
) -> RenderableAndLineAnnotations:
  """Builds a foldable tree node with path buttons.

  Args:
    prefix: Contents of the first line, before the children. Should not contain
      any other foldables. Usually ends with an opening paren/bracket, e.g.
      "SomeClass("
    children: Sequence of children of this node, which should each be rendered
      on their own line.
    suffix: Contents of the last line, after the children. Usually a closing
      paren/bracket for `prefix`.
    comma_separated: Whether to insert commas between children.
    force_trailing_comma: Whether to always insert a trailing comma after the
      last child.
    path: Keypath to this node from the root. If provided, copy-path buttons
      will be added.
    background_color: Optional background and border color for this node.
    background_pattern: Optional background pattern as a CSS "image". If
      provided, `background_color` must also be provided, and will be used as
      the border for the pattern.
    first_line_annotation: An annotation for the first line of the node when it
      is expanded.
    expand_state: Initial expand state for the foldable.

  Returns:
    A new renderable part, possibly with a copy button annotation, for use
    in part of a rendered treescope tree.
  """
  if not children:
    return build_one_line_tree_node(
        line=basic_parts.siblings(prefix, suffix),
        path=path,
        background_color=background_color,
    )

  maybe_copy_button = build_copy_button(path)

  if isinstance(prefix, str):
    prefix = basic_parts.Text(prefix)

  if isinstance(suffix, str):
    suffix = basic_parts.Text(suffix)

  if background_pattern is not None:
    if background_color is None:
      raise ValueError(
          "background_color must be provided if background_pattern is"
      )

    def wrap_block(block):
      return common_styles.WithBlockPattern(
          block, color=background_color, pattern=background_pattern
      )

    wrap_topline = common_styles.PatternedTopLineSpanGroup
    wrap_bottomline = common_styles.PatternedBottomLineSpanGroup
    indented_child_class = common_styles.ColoredBorderIndentedChildren

  elif background_color is not None and background_color != "transparent":

    def wrap_block(block):
      return common_styles.WithBlockColor(block, color=background_color)

    wrap_topline = common_styles.ColoredTopLineSpanGroup
    wrap_bottomline = common_styles.ColoredBottomLineSpanGroup
    indented_child_class = common_styles.ColoredBorderIndentedChildren

  else:
    wrap_block = lambda rendering: rendering
    wrap_topline = lambda rendering: rendering
    wrap_bottomline = lambda rendering: rendering
    indented_child_class = basic_parts.IndentedChildren

  if first_line_annotation is not None:
    maybe_first_line_annotation = basic_parts.FoldCondition(
        expanded=first_line_annotation
    )
  else:
    maybe_first_line_annotation = basic_parts.EmptyPart()

  return RenderableAndLineAnnotations(
      renderable=wrap_block(
          foldable_impl.FoldableTreeNodeImpl(
              label=wrap_topline(prefix),
              contents=basic_parts.siblings(
                  maybe_copy_button,
                  maybe_first_line_annotation,
                  indented_child_class.build(
                      children,
                      comma_separated=comma_separated,
                      force_trailing_comma=force_trailing_comma,
                  ),
                  wrap_bottomline(suffix),
              ),
              expand_state=expand_state,
          )
      ),
      annotations=maybe_copy_button,
  )


def maybe_qualified_type_name(ty: type[Any]) -> RenderableTreePart:
  """Formats the name of a type so that it is qualified in roundtrip mode.

  Args:
    ty: A type object to render.

  Returns:
    A tree part that renders as a fully-qualified name in roundtrip mode, or as
    the simple class name otherwise. Module names will be either inferred from
    the type's definition or looked up using `canonical_aliases`.
  """
  class_name = ty.__name__

  alias = canonical_aliases.lookup_alias(
      ty, allow_outdated=True, allow_relative=True
  )
  if alias:
    access_path = str(alias)
  else:
    access_path = f"<unknown>.{class_name}"

  if access_path.endswith(class_name):
    return basic_parts.siblings(
        basic_parts.RoundtripCondition(
            roundtrip=common_styles.QualifiedTypeNameSpanGroup(
                basic_parts.Text(access_path.removesuffix(class_name))
            )
        ),
        basic_parts.Text(class_name),
    )
  else:
    return basic_parts.RoundtripCondition(
        roundtrip=basic_parts.Text(access_path),
        not_roundtrip=basic_parts.Text(class_name),
    )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for inspecting objects."""


from __future__ import annotations
import abc
import types
from typing import Any, Callable


def safely_get_real_method(
    obj: Any, method_name: str
) -> Callable[..., Any] | None:
  """Safely retrieves a method directly implemented on a type.

  This function can be used to safely retrieve a "real" method from an object,
  e.g. a method that is actually defined on the object or a superclass. This
  can be used to avoid accidentally trying to call special methods on proxy
  objects or other objects that override `__getattr__`.

  Args:
    obj: The object to retrieve the method from.
    method_name: The name of the method to retrieve.

  Returns:
    The method if it is found, or None if it is not found.
  """
  if not hasattr(type(obj), method_name) or not hasattr(obj, method_name):
    return None

  try:
    retrieved = getattr(obj, method_name)
    if not isinstance(retrieved, types.MethodType):
      return None
    return retrieved
  except Exception:  # pylint: disable=broad-exception-caught
    return None


class HasReprHtml(abc.ABC):
  """Abstract base class for rich-display objects in IPython."""

  @abc.abstractmethod
  def _repr_html_(self) -> str | tuple[str, Any]:
    """Returns a rich HTML representation of an object."""

  @classmethod
  def __subclasshook__(cls, subclass, /):
    """Checks if a class is a subclass of HasReprHtml."""
    return hasattr(subclass, '_repr_html_') and callable(subclass._repr_html_)  # pylint: disable=protected-access


def to_html(node: Any) -> str | None:
  """Extracts a rich HTML representation of node using _repr_html_."""
  repr_html_method = safely_get_real_method(node, '_repr_html_')
  if repr_html_method is None:
    return None
  html_for_node_and_maybe_metadata = repr_html_method()
  if isinstance(html_for_node_and_maybe_metadata, tuple):
    html_for_node, _ = html_for_node_and_maybe_metadata
  else:
    html_for_node = html_for_node_and_maybe_metadata
  return html_for_node



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal implementation of array visualizer."""


from __future__ import annotations

import base64
import dataclasses
import io
import json
import os
from typing import Any, Literal, Sequence

import numpy as np
from treescope import ndarray_adapters
from treescope._internal import html_escaping
from treescope._internal.parts import basic_parts
from treescope._internal.parts import part_interface


AxisName = Any

PositionalAxisInfo = ndarray_adapters.PositionalAxisInfo
NamedPositionlessAxisInfo = ndarray_adapters.NamedPositionlessAxisInfo
NamedPositionalAxisInfo = ndarray_adapters.NamedPositionalAxisInfo
AxisInfo = ndarray_adapters.AxisInfo

ArrayInRegistry = Any


def load_arrayvis_javascript() -> str:
  """Loads the contents of `arrayvis.js` from the Python package.

  Returns:
    Source code for arrayviz.
  """
  filepath = __file__
  if filepath is None:
    raise ValueError("Could not find the path to arrayviz.js!")

  # Look for the resource relative to the current module's filesystem path.
  base = filepath.removesuffix("arrayviz_impl.py")
  load_path = os.path.join(base, "js", "arrayviz.js")

  with open(load_path, "r", encoding="utf-8") as f:
    return f.read()


def html_setup() -> (
    set[part_interface.CSSStyleRule | part_interface.JavaScriptDefn]
):
  """Builds the setup HTML that should be included in any arrayviz output."""
  arrayviz_src = html_escaping.heuristic_strip_javascript_comments(
      load_arrayvis_javascript()
  )
  return {
      part_interface.CSSStyleRule(html_escaping.without_repeated_whitespace("""
        .arrayviz_container {
            white-space: normal;
        }
        .arrayviz_container .info {
            font-family: monospace;
            color: #aaaaaa;
            margin-bottom: 0.25em;
            white-space: pre;
        }
        .arrayviz_container .info input[type="range"] {
            vertical-align: middle;
            filter: grayscale(1) opacity(0.5);
        }
        .arrayviz_container .info input[type="range"]:hover {
            filter: grayscale(0.5);
        }
        .arrayviz_container .info input[type="number"]:not(:focus) {
            border-radius: 3px;
        }
        .arrayviz_container .info input[type="number"]:not(:focus):not(:hover) {
            color: #777777;
            border: 1px solid #777777;
        }
        .arrayviz_container .info.sliders {
            white-space: pre;
        }
        .arrayviz_container .hovertip {
            display: none;
            position: absolute;
            background-color: white;
            border: 1px solid black;
            padding: 0.25ch;
            pointer-events: none;
            width: fit-content;
            overflow: visible;
            white-space: pre;
            z-index: 1000;
        }
        .arrayviz_container .hoverbox {
            display: none;
            position: absolute;
            box-shadow: 0 0 0 1px black, 0 0 0 2px white;
            pointer-events: none;
            z-index: 900;
        }
        .arrayviz_container .clickdata {
            white-space: pre;
        }
        .arrayviz_container .loading_message {
            color: #aaaaaa;
        }
      """)),
      part_interface.JavaScriptDefn(
          arrayviz_src + " this.getRootNode().host.defns.arrayviz = arrayviz;"
      ),
  }


def render_array_data_to_html(
    array_data: np.ndarray,
    valid_mask: np.ndarray,
    column_axes: Sequence[int],
    row_axes: Sequence[int],
    slider_axes: Sequence[int],
    axis_labels: list[str],
    vmin: float,
    vmax: float,
    cmap_type: Literal["continuous", "palette_index", "digitbox"],
    cmap_data: list[tuple[int, int, int]],
    info: str = "",
    formatting_instructions: list[dict[str, Any]] | None = None,
    dynamic_continuous_cmap: bool = False,
    raw_min_abs: float | None = None,
    raw_max_abs: float | None = None,
    pixels_per_cell: int | float = 7,
) -> str:
  """Helper to render an array to HTML by passing arguments to javascript.

  Args:
    array_data: Array data to render.
    valid_mask: Mask array, of same shape as array_data, that is True for items
      we should render.
    column_axes: Axes (by index into `array_data`) to arrange as columns,
      ordered from outermost group to innermost group.
    row_axes: Axes (by index into `array_data`) to arrange as rows, ordered from
      outermost group to innermost group.
    slider_axes: Axes to bind to sliders.
    axis_labels: Labels for each axis.
    vmin: Minimum for the colormap.
    vmax: Maximum for the colormap.
    cmap_type: Type of colormap (see `render_array`)
    cmap_data: Data for the colormap, as a sequence of RGB triples.
    info: Info for the plot.
    formatting_instructions: Formatting instructions for values on mouse hover
      or click. These will be interpreted by `formatValueAndIndices` on the
      JavaScript side. Can assume each axis is named "a0", "a1", etc. when
      running in JavaScript.
    dynamic_continuous_cmap: Whether to dynamically adjust the colormap during
      rendering.
    raw_min_abs: Minimum absolute value of the array, for dynamic remapping.
    raw_max_abs: Maximum absolute value of the array, for dynamic remapping.
    pixels_per_cell: The initial number of pixels per cell when rendering.

  Returns:
    HTML source for an arrayviz rendering.
  """
  assert len(array_data.shape) == len(axis_labels)
  assert len(valid_mask.shape) == len(axis_labels)

  if formatting_instructions is None:
    formatting_instructions = [{"type": "value"}]

  # Compute strides for each axis. We refer to each axis as "a0", "a1", etc
  # across the JavaScript boundary.
  stride = 1
  strides = {}
  for i, axis_size in reversed(list(enumerate(array_data.shape))):
    strides[f"a{i}"] = stride
    stride *= axis_size

  if cmap_type == "continuous":
    converted_array_data = array_data.astype(np.float32)
    array_dtype = "float32"
  else:
    converted_array_data = array_data.astype(np.int32)
    array_dtype = "int32"

  def axis_spec_arg(i):
    return {
        "name": f"a{i}",
        "label": axis_labels[i],
        "start": 0,
        "end": array_data.shape[i],
    }

  x_axis_specs_arg = []
  for axis in column_axes:
    x_axis_specs_arg.append(axis_spec_arg(axis))

  y_axis_specs_arg = []
  for axis in row_axes:
    y_axis_specs_arg.append(axis_spec_arg(axis))

  sliced_axis_specs_arg = []
  for axis in slider_axes:
    sliced_axis_specs_arg.append(axis_spec_arg(axis))

  args_json = json.dumps({
      "info": info,
      "arrayBase64": base64.b64encode(converted_array_data.tobytes()).decode(
          "ascii"
      ),
      "arrayDtype": array_dtype,
      "validMaskBase64": base64.b64encode(
          valid_mask.astype(np.uint8).tobytes()
      ).decode("ascii"),
      "dataStrides": strides,
      "xAxisSpecs": x_axis_specs_arg,
      "yAxisSpecs": y_axis_specs_arg,
      "slicedAxisSpecs": sliced_axis_specs_arg,
      "colormapConfig": {
          "type": cmap_type,
          "min": vmin,
          "max": vmax,
          "dynamic": dynamic_continuous_cmap,
          "rawMinAbs": raw_min_abs,
          "rawMaxAbs": raw_max_abs,
          "cmapData": cmap_data,
      },
      "pixelsPerCell": pixels_per_cell,
      "valueFormattingInstructions": formatting_instructions,
  })
  # Note: We need to save the parent of the treescope-run-here element first,
  # because it will be removed before the runSoon callback executes.
  inner_fn = html_escaping.without_repeated_whitespace("""
    const parent = this.parentNode;
    const defns = this.getRootNode().host.defns;
    defns.runSoon(() => {
        const tpl = parent.querySelector('template.deferred_args');
        const config = JSON.parse(
            tpl.content.querySelector('script').textContent
        );
        tpl.remove();
        defns.arrayviz.buildArrayvizFigure(parent, config);
    });
  """)
  src = (
      '<div class="arrayviz_container">'
      '<span class="loading_message">Rendering array...</span>'
      f'<treescope-run-here><script type="application/octet-stream">{inner_fn}'
      "</script></treescope-run-here>"
      '<template class="deferred_args">'
      f'<script type="application/json">{args_json}</script></template></div>'
  )
  return src


def infer_rows_and_columns(
    all_axes: Sequence[AxisInfo],
    known_rows: Sequence[AxisInfo] = (),
    known_columns: Sequence[AxisInfo] = (),
    edge_items_per_axis: tuple[int | None, ...] | None = None,
) -> tuple[list[AxisInfo], list[AxisInfo]]:
  """Infers an ordered assignment of axis indices or names to rows and columns.

  The unassigned axes are sorted by size and then assigned to rows and columns
  to try to balance the total number of elements along the row and column axes.
  This currently uses a greedy algorithm with an adjustment to try to keep
  columns longer than rows, except when there are exactly two axes and both are
  positional, in which case it lays out axis 0 as the rows and axis 1 as the
  columns.

  Axes with logical positions are sorted before axes with only names
  (in reverse order, so that later axes are rendered on the inside). Axes with
  names only appear afterward, with explicitly-assigned ones before unassigned
  ones.

  Args:
    all_axes: Sequence of axis infos in the array that should be assigned.
    known_rows: Sequence of axis indices or names that must map to rows.
    known_columns: Sequence of axis indices or names that must map to columns.
    edge_items_per_axis: Optional edge items specification, determining
      truncated size of each axis. Must match the ordering of `all_axes`.

  Returns:
    Tuple (rows, columns) of assignments, which consist of `known_rows` and
    `known_columns` followed by the remaining unassigned axes in a balanced
    order.
  """
  if edge_items_per_axis is None:
    edge_items_per_axis = (None,) * len(all_axes)

  if not known_rows and not known_columns and len(all_axes) == 2:
    ax_a, ax_b = all_axes
    if (
        isinstance(ax_a, PositionalAxisInfo)
        and isinstance(ax_b, PositionalAxisInfo)
        and {ax_a.axis_logical_index, ax_b.axis_logical_index} == {0, 1}
    ):
      # Two-dimensional positional array. Always do rows then columns.
      if ax_a.axis_logical_index == 0:
        return ([ax_a], [ax_b])
      else:
        return ([ax_b], [ax_a])

  truncated_sizes = {
      ax: ax.size if edge_items is None else 2 * edge_items + 1
      for ax, edge_items in zip(all_axes, edge_items_per_axis)
  }
  unassigned = [
      ax for ax in all_axes if ax not in known_rows and ax not in known_columns
  ]

  # Sort by size descending, so that we make the most important layout decisions
  # first.
  unassigned = sorted(
      unassigned, key=lambda ax: (truncated_sizes[ax], ax.size), reverse=True
  )

  # Compute the total size every axis would have if we assigned them to the
  # same axis.
  unassigned_size = np.prod([truncated_sizes[ax] for ax in unassigned])

  rows = list(known_rows)
  row_size = np.prod([truncated_sizes[ax] for ax in rows])
  columns = list(known_columns)
  column_size = np.prod([truncated_sizes[ax] for ax in columns])

  for ax in unassigned:
    axis_size = truncated_sizes[ax]
    unassigned_size = unassigned_size // axis_size
    if row_size * axis_size > column_size * unassigned_size:
      # If we assign this to the row axis, we'll end up with a visualization
      # with more rows than columns regardless of what we do later, which can
      # waste screen space. Assign to columns instead.
      columns.append(ax)
      column_size *= axis_size
    else:
      # Assign to the row axis. We'll assign columns later.
      rows.append(ax)
      row_size *= axis_size

  # The specific ordering of axes along the rows and the columns is somewhat
  # arbitrary. Re-order each so that explicitly requested axes are first, then
  # unassigned positional axes in reverse position order, then unassigned named
  # axes.
  def ax_sort_key(ax: AxisInfo):
    if ax not in unassigned:
      return (0,)
    elif isinstance(ax, PositionalAxisInfo | NamedPositionalAxisInfo):
      return (1, -ax.axis_logical_index)
    else:
      return (2,)

  return sorted(rows, key=ax_sort_key), sorted(columns, key=ax_sort_key)


def infer_vmin_vmax(
    array: np.ndarray,
    mask: np.ndarray,
    vmin: float | None,
    vmax: float | None,
    around_zero: bool,
    trim_outliers: bool,
) -> tuple[float, float]:
  """Infer reasonable lower and upper colormap bounds from an array."""
  inferring_both_bounds = vmax is None and vmin is None
  finite_mask = np.logical_and(np.isfinite(array), mask)
  if vmax is None:
    if around_zero:
      if vmin is not None:
        vmax = -vmin  # pylint: disable=invalid-unary-operand-type
      else:
        vmax = np.max(np.where(finite_mask, np.abs(array), 0))
    else:
      vmax = np.max(np.where(finite_mask, array, -np.inf))

  assert vmax is not None

  if vmin is None:
    if around_zero:
      vmin = -vmax  # pylint: disable=invalid-unary-operand-type
    else:
      vmin = np.min(np.where(finite_mask, array, np.inf))

  if inferring_both_bounds and trim_outliers:
    if around_zero:
      center = 0
    else:
      center = np.nanmean(np.where(finite_mask, array, np.nan))
      center = np.where(np.isfinite(center), center, 0.0)

    second_moment = np.nanmean(
        np.where(finite_mask, np.square(array - center), np.nan)
    )
    sigma = np.where(
        np.isfinite(second_moment), np.sqrt(second_moment), vmax - vmin
    )

    vmin_limit = center - 3 * sigma
    vmin = np.maximum(vmin, vmin_limit)
    vmax_limit = center + 3 * sigma
    vmax = np.minimum(vmax, vmax_limit)

  return vmin, vmax


def infer_abs_min_max(
    array: np.ndarray, mask: np.ndarray
) -> tuple[float, float]:
  """Infer smallest and largest absolute values in array."""
  finite_mask = np.logical_and(np.isfinite(array), mask)
  absmin = np.min(
      np.where(np.logical_and(finite_mask, array != 0), np.abs(array), np.inf)
  )
  absmin = np.where(np.isinf(absmin), 0.0, absmin)
  absmax = np.max(np.where(finite_mask, np.abs(array), -np.inf))
  absmax = np.where(np.isinf(absmax), 0.0, absmax)
  return absmin, absmax


def infer_balanced_truncation(
    shape: Sequence[int],
    maximum_size: int,
    cutoff_size_per_axis: int,
    minimum_edge_items: int,
    doubling_bonus: float = 10.0,
) -> tuple[int | None, ...]:
  """Infers a balanced truncation from a shape.

  This function computes a set of truncation sizes for each axis of the array
  such that it obeys the constraints about array and axis sizes, while also
  keeping the relative proportions of the array consistent (e.g. we keep more
  elements along axes that were originally longer). This means that the aspect
  ratio of the truncated array will still resemble the aspect ratio of the
  original array.

  To avoid very-unbalanced renderings and truncate longer axes more than short
  ones, this function truncates based on the square-root of the axis size by
  default.

  Args:
    shape: The shape of the array we are truncating.
    maximum_size: Maximum number of elements of an array to show. Arrays larger
      than this will be truncated along one or more axes.
    cutoff_size_per_axis: Maximum number of elements of each individual axis to
      show without truncation. Any axis longer than this will be truncated, with
      their visual size increasing logarithmically with the true axis size
      beyond this point.
    minimum_edge_items: How many values to keep along each axis for truncated
      arrays. We may keep more than this up to the budget of maximum_size.
    doubling_bonus: Number of elements to add to each axis each time it doubles
      beyond `cutoff_size_per_axis`. Used to make longer axes appear visually
      longer while still keeping them a reasonable size.

  Returns:
    A tuple of edge sizes. Each element corresponds to an axis in `shape`,
    and is either `None` (for no truncation) or an integer (corresponding to
    the number of elements to keep at the beginning and at the end).
  """
  shape_arr = np.array(list(shape))
  remaining_elements_to_divide = maximum_size
  edge_items_per_axis = {}
  # Order our shape from smallest to largest, since the smallest axes will
  # require the least amount of truncation and will have the most stringent
  # constraints.
  sorted_axes = np.argsort(shape_arr)
  sorted_shape = shape_arr[sorted_axes]

  # Figure out maximum sizes based on the cutoff
  cutoff_adjusted_maximum_sizes = np.where(
      sorted_shape <= cutoff_size_per_axis,
      sorted_shape,
      cutoff_size_per_axis
      + doubling_bonus * np.log2(sorted_shape / cutoff_size_per_axis),
  )

  # Suppose we want to make a scaled version of the array with relative
  # axis sizes
  #   s0, s1, s2, ...
  # The total size is then
  #   size = (c * s0) * (c * s1) * (c * s2) * ...
  #   log(size) = ndim * log(c) + [ log s0 + log s1 + log s2 + ... ]
  # If we have a known final size we want to reach, we can solve for c as
  #   c = exp( (log size - [ log s0 + log s1 + log s2 + ... ]) / ndim )
  axis_proportions = np.sqrt(sorted_shape)
  log_axis_proportions = np.log(axis_proportions)
  for i in range(len(sorted_axes)):
    original_axis = sorted_axes[i]
    size = shape_arr[original_axis]
    # If we truncated this axis and every axis after it proportional to
    # their weights, how small of an axis size would we need for this
    # axis?
    log_c = (
        np.log(remaining_elements_to_divide) - np.sum(log_axis_proportions[i:])
    ) / (len(shape) - i)
    soft_limit_for_this_axis = np.exp(log_c + log_axis_proportions[i])
    cutoff_limit_for_this_axis = np.floor(
        np.minimum(
            soft_limit_for_this_axis,
            cutoff_adjusted_maximum_sizes[i],
        )
    )
    if size <= 2 * minimum_edge_items + 1 or size <= cutoff_limit_for_this_axis:
      # If this axis is already smaller than the minimum size it would have
      # after truncation, there's no reason to truncate it.
      # But pretend we did, so that other axes still grow monotonically if
      # their axis sizes increase.
      remaining_elements_to_divide = (
          remaining_elements_to_divide / soft_limit_for_this_axis
      )
      edge_items_per_axis[original_axis] = None
    elif cutoff_limit_for_this_axis < 2 * minimum_edge_items + 1:
      # If this axis is big enough to truncate, but our naive target size is
      # smaller than the minimum allowed truncation, we should truncate it
      # to the minimum size allowed instead.
      edge_items_per_axis[original_axis] = minimum_edge_items
      remaining_elements_to_divide = remaining_elements_to_divide / (
          2 * minimum_edge_items + 1
      )
    else:
      # Otherwise, truncate it and all remaining axes based on our target
      # truncations.
      for j in range(i, len(sorted_axes)):
        visual_size = np.floor(
            np.minimum(
                np.exp(log_c + log_axis_proportions[j]),
                cutoff_adjusted_maximum_sizes[j],
            )
        )
        edge_items_per_axis[sorted_axes[j]] = int(visual_size // 2)
      break

  return tuple(
      edge_items_per_axis[orig_axis] for orig_axis in range(len(shape))
  )


def compute_truncated_shape(
    shape: tuple[int, ...],
    edge_items: tuple[int | None, ...],
) -> tuple[int, ...]:
  """Computes the shape of a truncated array.

  This can be used to estimate the size of an array visualization after it has
  been truncated by `infer_balanced_truncation`.

  Args:
    shape: The original array shape.
    edge_items: Number of edge items to keep along each axis.

  Returns:
    The shape of the truncated array.
  """
  return tuple(
      orig if edge is None else 2 * edge + 1
      for orig, edge in zip(shape, edge_items)
  )


@dataclasses.dataclass(frozen=True)
class ArrayvizRendering(part_interface.RenderableTreePart):
  """A rendering of an array with Arrayviz.

  Attributes:
    html_src: HTML source for the rendering.
  """

  html_src: str

  def _compute_collapsed_width(self) -> int:
    return 80

  def _compute_newlines_in_expanded_parent(self) -> int:
    return 10

  def foldables_in_this_part(self) -> Sequence[part_interface.FoldableTreeNode]:
    return ()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset()

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    stream.write("<Arrayviz rendering>")

  def html_setup_parts(
      self, setup_context: part_interface.HtmlContextForSetup
  ) -> set[part_interface.CSSStyleRule | part_interface.JavaScriptDefn]:
    del setup_context
    return html_setup()

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    stream.write(self.html_src)


@dataclasses.dataclass(frozen=True)
class ArrayvizDigitboxRendering(ArrayvizRendering):
  """A rendering of a single digitbox with Arrayviz."""

  def _compute_collapsed_width(self) -> int:
    return 2

  def _compute_newlines_in_expanded_parent(self) -> int:
    return 1


@dataclasses.dataclass(frozen=True)
class ValueColoredTextbox(basic_parts.DeferringToChild):
  """A rendering of text with a colored background.

  Attributes:
    child: Child part to render.
    text_color: Color for the text.
    background_color: Color for the background, usually from a colormap.
    out_of_bounds: Whether this value was out of bounds of the colormap.
    value: Underlying float value that is being visualized. Rendered on hover.
  """

  child: part_interface.RenderableTreePart
  text_color: str
  background_color: str
  out_of_bounds: bool
  value: float

  def html_setup_parts(
      self, setup_context: part_interface.HtmlContextForSetup
  ) -> set[part_interface.CSSStyleRule | part_interface.JavaScriptDefn]:
    return (
        {
            part_interface.CSSStyleRule(
                html_escaping.without_repeated_whitespace("""
                    .arrayviz_textbox {
                        padding-left: 0.5ch;
                        padding-right: 0.5ch;
                        outline: 1px solid black;
                        position: relative;
                        display: inline-block;
                        font-family: monospace;
                        white-space: pre;
                        margin-top: 1px;
                        box-sizing: border-box;
                    }
                    .arrayviz_textbox.out_of_bounds {
                        outline: 3px double darkorange;
                    }
                    .arrayviz_textbox .value {
                        display: none;
                        position: absolute;
                        bottom: 110%;
                        left: 0;
                        overflow: visible;
                        color: black;
                        background-color: white;
                        font-size: 0.7em;
                    }
                    .arrayviz_textbox:hover .value {
                        display: block;
                    }
                    """)
            )
        }
        | self.child.html_setup_parts(setup_context)
    )

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    class_string = "arrayviz_textbox"
    if self.out_of_bounds:
      class_string += " out_of_bounds"
    bg_color = html_escaping.escape_html_attribute(self.background_color)
    text_color = html_escaping.escape_html_attribute(self.text_color)
    stream.write(
        f'<span class="{class_string}" style="background-color:{bg_color};'
        f' color:{text_color}">'
        f'<span class="value">{float(self.value):.4g}</span>'
    )
    self.child.render_to_html(
        stream,
        at_beginning_of_line=False,
        render_context=render_context,
    )
    stream.write("</span>")



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines some classes for use in rendering tests.

This is in a separate module so we can test treescope's handling of module
names during qualified name lookup.
"""

from __future__ import annotations

import dataclasses
import enum
import typing
from typing import Any

import jax
import torch
import treescope


class MyTestEnum(enum.Enum):
  FOO = 1
  BAR = 2


@dataclasses.dataclass
class DataclassWithOneChild:
  foo: Any


@dataclasses.dataclass
class DataclassWithTwoChildren:
  foo: Any
  bar: Any


@dataclasses.dataclass(frozen=True)
class EmptyDataclass:
  pass


class SomeNamedtupleClass(typing.NamedTuple):
  foo: Any
  bar: Any


class SomeOuterClass:

  @dataclasses.dataclass
  class NestedClass:
    foo: Any


def make_class_with_weird_qualname():
  @dataclasses.dataclass
  class ClassDefinedInAFunction:  # pylint: disable=redefined-outer-name
    foo: Any

  return ClassDefinedInAFunction


ClassDefinedInAFunction = make_class_with_weird_qualname()


class RedefinedClass:
  """A class that will later be redefined."""

  pass


OriginalRedefinedClass = RedefinedClass


class RedefinedClass:  # pylint: disable=function-redefined
  """The redefined class; no longer the same as `_OriginalRedefinedClass`."""

  pass


class _PrivateClass:

  def some_function(self):
    pass


def _private_function():
  pass


class SomeFunctionLikeWrapper:
  func: Any

  def __init__(self, func):
    self.func = func

  def __call__(self, *args, **kwargs):
    return self.func(*args, **kwargs)

  @property
  def __wrapped__(self):
    return self.func


@SomeFunctionLikeWrapper
def wrapped_function():
  pass


immutable_constant = (1, 2, 3)
mutable_constant = [1, 2, 3]


class SomethingCallable:

  def __call__(self, value: int) -> int:
    return value + 1


# pytype is confused about the dataclass transform here
some_callable_block = SomethingCallable()


@dataclasses.dataclass(frozen=True)
class FrozenDataclassKey:
  name: str


@jax.tree_util.register_pytree_with_keys_class
class UnknownPytreeNode:
  """A Pytree node treescope doesn't know."""

  def __init__(self, x, y):
    self.x = x
    self.y = y

  def __repr__(self):
    return f"<custom repr for UnknownPytreeNode: x={self.x}, y={self.y}>"

  def tree_flatten_with_keys(self):
    return (
        ((FrozenDataclassKey("x"), self.x), ("string_key", self.y)),
        "example_pytree_aux_data",
    )

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del aux_data
    return cls(*children)


class UnknownObjectWithBuiltinRepr:
  pass


class UnknownObjectWithOneLineRepr:

  def __repr__(self):
    return "<custom repr for UnknownObjectWithOneLineRepr>"


class UnknownObjectWithMultiLineRepr:

  def __repr__(self):
    return "<custom repr\n  for\n  UnknownObjectWithMultiLineRepr\n>"


class UnknownObjectWithBadMultiLineRepr:

  def __repr__(self):
    return "Non-idiomatic\nmultiline\nobject"


class ObjectWithCustomHandler:

  def __treescope_repr__(self, path, subtree_renderer):
    del subtree_renderer
    return treescope.rendering_parts.text(
        f"<ObjectWithCustomHandler custom rendering! Path: {repr(path)}>"
    )


class ObjectWithCustomHandlerThatThrows:

  def __treescope_repr__(self, path, subtree_renderer):
    del path, subtree_renderer
    raise RuntimeError("Simulated treescope_repr failure!")

  def __repr__(self):
    return "<Fallback repr for ObjectWithCustomHandlerThatThrows>"


class ObjectWithReprThatThrows:

  def __repr__(self):
    raise RuntimeError("Simulated repr failure!")


class ObjectWithCustomHandlerThatThrowsDeferred:

  def __treescope_repr__(self, path, subtree_renderer):
    del path, subtree_renderer

    def _internal_main_thunk(layout_decision):
      del layout_decision
      raise RuntimeError("Simulated deferred treescope_repr failure!")

    return treescope.lowering.maybe_defer_rendering(
        main_thunk=_internal_main_thunk,
        placeholder_thunk=lambda: treescope.rendering_parts.text(
            "<deferred placeholder>"
        ),
    )


class SomePyTorchModule(torch.nn.Module):
  """A basic PyTorch module to test rendering."""

  def __init__(self):
    super().__init__()
    # Attributes
    self.attr_one = 123
    self.attr_two = "abc"
    # Child modules
    self.linear = torch.nn.Linear(10, 10)
    self.mod_list = torch.nn.ModuleList(
        [torch.nn.LayerNorm(10), torch.nn.SiLU()]
    )
    # Parameters
    self.foo = torch.nn.Parameter(torch.ones(5))
    # Buffers
    self.register_buffer("bar", torch.zeros(5))

  @classmethod
  def build(cls):
    torch.random.manual_seed(1234)
    return cls()



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Postprocessor to annotate repetitions of mutable values.

This can be useful for figuring out when multiple mutable objects in a tree are
all references to the same object.
"""

import contextlib
import dataclasses
import io
from typing import Any, Optional, Sequence

from treescope import context
from treescope import renderers
from treescope import rendering_parts
from treescope import type_registries
from treescope._internal import html_escaping
from treescope._internal.parts import basic_parts
from treescope._internal.parts import part_interface


@dataclasses.dataclass
class _SharedObjectTracker:
  """Helper object to track IDs we've seen before.

  This object keeps track of nodes we've encountered while rendering the root
  object, so that we can detect when the same object is encountered in multiple
  places. To ensure that we don't have false positives, we also store the nodes
  themselves until the rendering finishes. (Usually, rendered Python objects
  will be kept alive due to being part of the rendered tree, but some handlers
  may render temporary objects that are discarded, at which point their object
  IDs are sometimes re-used by other temporary objects.)

  Attributes:
    seen_at_least_once: Map with node IDs as keys and nodes as values,
      containing nodes we've seen at least once.
    seen_more_than_once: Set of node IDs we've seen more than once.
  """

  seen_at_least_once: dict[int, Any]
  seen_more_than_once: set[int]


_shared_object_ids_seen: context.ContextualValue[
    Optional[_SharedObjectTracker]
] = context.ContextualValue(
    module=__name__, qualname="_shared_object_ids_seen", initial_value=None
)
"""A collection of IDs of objects that we have already rendered.

This is used to detect "unsafe" sharing of objects.

Normally references to the same object are ignored by JAX, and doing any
PyTree manipulation destroys the sharing. However, leaves of JAX PyTrees
may still involve mutable-object sharing, and we want to be able to
use Treescope to visualize this kind of structure as well. We also
use this to avoid infinite recursion if a node in the tree contains
itself.

This context manager should only be used by code in this module.
"""


class SharedWarningLabel(basic_parts.BaseSpanGroup):
  """A comment identifying this shared Python object."""

  def _span_css_class(self) -> str:
    return "shared_warning_label"

  def _span_css_rule(
      self, setup_context: part_interface.HtmlContextForSetup
  ) -> part_interface.CSSStyleRule:
    del setup_context
    return part_interface.CSSStyleRule(
        html_escaping.without_repeated_whitespace("""
            .shared_warning_label
            {
                color: #ffac13;
            }
        """)
    )


@dataclasses.dataclass(frozen=False)
class DynamicSharedCheck(rendering_parts.RenderableTreePart):
  """Dynamic group that renders its child only if a node is shared.

  This node is used to apply special rendering to nodes that are encountered in
  multiple places in the same input. It works by holding a particular node ID
  as well as a reference to a set of node IDs that have been seen more than
  once. When rendered, it checks if its node ID has been seen more than once
  and renders differently if so.

  Note that we might have only seen this object once at the time the shared
  warning object is constructed, even if eventually we will see it again. This
  means the `seen_more_than_once` attribute has to be (a reference to) a
  *mutable set* modified elsewhere.

  Shared markers always act like empty parts for layout decisions.

  Attributes:
    if_shared: Child to render only if the node is shared.
    node_id: Node ID of the node we are rendering. Used for the warning and also
      looked up in `seen_more_than_once`.
    seen_more_than_once: Reference to an externally-maintained set of node IDs
      we've seen more than once. Usually, this will be the same set from the
      active `_shared_object_ids_seen` context.
  """

  if_shared: rendering_parts.RenderableTreePart
  node_id: int
  seen_more_than_once: set[int]

  def _compute_collapsed_width(self) -> int:
    return 0

  def _compute_newlines_in_expanded_parent(self) -> int:
    return 0

  def foldables_in_this_part(self) -> Sequence[part_interface.FoldableTreeNode]:
    return []

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset()

  def html_setup_parts(
      self, setup_context: part_interface.HtmlContextForSetup
  ) -> set[part_interface.CSSStyleRule | part_interface.JavaScriptDefn]:
    return self.if_shared.html_setup_parts(setup_context)

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    if self.node_id in self.seen_more_than_once:
      self.if_shared.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    if self.node_id in self.seen_more_than_once:
      self.if_shared.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )


@dataclasses.dataclass(frozen=False)
class WithDynamicSharedPip(basic_parts.DeferringToChild):
  """Dynamic group that adds an orange marker to its child if it is shared.

  This node is used to apply special rendering to nodes that are encountered in
  multiple places in the same input. It works by holding a particular node ID
  as well as a reference to a set of node IDs that have been seen more than
  once. When rendered, it checks if its node ID has been seen more than once
  and renders differently if so.

  Note that we might have only seen this object once at the time the shared
  warning object is constructed, even if eventually we will see it again. This
  means the `seen_more_than_once` attribute has to be (a reference to) a
  *mutable set* modified elsewhere.

  Shared markers always act like their child. They also do not disrupt children
  drawing themselves first.

  Attributes:
    child: Child to render.
    node_id: Node ID of the node we are rendering. Used for the warning and also
      looked up in `seen_more_than_once`.
    seen_more_than_once: Reference to an externally-maintained set of node IDs
      we've seen more than once. Usually, this will be the same set from the
      active `_shared_object_ids_seen` context.
  """

  child: rendering_parts.RenderableTreePart
  node_id: int
  seen_more_than_once: set[int]

  def html_setup_parts(
      self, setup_context: part_interface.HtmlContextForSetup
  ) -> set[part_interface.CSSStyleRule | part_interface.JavaScriptDefn]:
    return self.child.html_setup_parts(setup_context) | {
        part_interface.CSSStyleRule(
            html_escaping.without_repeated_whitespace(f"""
                .shared_warning_pip
                {{
                    padding-right: 1ch;
                    margin-right: -0.5ch;
                    background: linear-gradient(
                        135deg, orange 0 0.6ch, transparent 0.6ch
                    );
                }}
                .shared_warning_pip.is_first_on_line:not(
                    {setup_context.collapsed_selector} *
                )
                {{
                    margin-left: -0.5ch;
                }}
            """)
        )
    }

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    if self.node_id in self.seen_more_than_once:
      if at_beginning_of_line:
        stream.write(
            "<span class='shared_warning_pip is_first_on_line'></span>"
        )
      else:
        stream.write("<span class='shared_warning_pip'></span>")
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )


def setup_shared_value_context() -> contextlib.AbstractContextManager[None]:
  """Returns a context manager for the shared value context.

  Within this context manager, `_shared_object_ids_seen` will refer
  to a consistent tracker. This tracker can then be used to check for repeated
  appearances of the same mutable object.

  This should be included in the `context_builders` argument to any renderer
  that checks for shared values.
  """
  return _shared_object_ids_seen.set_scoped(_SharedObjectTracker({}, set()))


def _is_safe_to_share(node: Any) -> bool:
  """Returns whether the given node is immutable."""
  # According to the Python data model, "If a class defines mutable objects and
  # implements an __eq__() method, it should not implement __hash__()". So, if
  # we find an object that implements __eq__ and __hash__, we can generally
  # assume it is immutable.
  return (
      type(node).__hash__ is not None
      and type(node).__hash__ is not object.__hash__
      and type(node).__eq__ is not object.__eq__
  ) or type_registries.lookup_immutability_for_type(type(node))


def check_for_shared_values(
    node: Any,
    path: str | None,
    node_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  # pylint: disable=g-doc-args,g-doc-return-or-yield
  """Wrapper hook to check for and annotate shared values.

  This wrapper should only be used by renderers that also include
  `build_styles_for_shared_values` in their HTML configuration and
  `setup_shared_value_context` in their context builders.

  Args:
    node: The node that has been rendered
    path: Optionally, a path to this node as a string.
    node_renderer: The inner renderer for this node. This should be used to
      render `node` itself into HTML tags.

  Returns:
    A possibly-modified representation of this object.

  Raises:
    RuntimeError: If called outside of the context constructed via
      setup_shared_value_context.
  """
  # pylint: enable=g-doc-args,g-doc-return-or-yield
  shared_object_tracker = _shared_object_ids_seen.get()
  if shared_object_tracker is None:
    raise RuntimeError(
        "`check_for_shared_values` should only be called in a shared value"
        " context! Make sure the current treescope renderer has"
        " `setup_shared_value_context` in its `context_builders`."
    )

  # Use object identity to track shared references and loops.
  node_id = id(node)

  # For types that we know are immutable, it's not necessary to render shared
  # references in a special way.
  safe_to_share = _is_safe_to_share(node)

  # Render the node normally.
  rendering = node_renderer(node, path)

  if not safe_to_share:
    # Mark this as possibly shared.
    if node_id in shared_object_tracker.seen_at_least_once:
      shared_object_tracker.seen_more_than_once.add(node_id)
    else:
      shared_object_tracker.seen_at_least_once[node_id] = node

    # Wrap it in a shared value wrapper; this will check to see if the same
    # node was seen more than once, and add an annotation if so.
    return rendering_parts.RenderableAndLineAnnotations(
        renderable=WithDynamicSharedPip(
            rendering.renderable,
            node_id=node_id,
            seen_more_than_once=shared_object_tracker.seen_more_than_once,
        ),
        annotations=rendering_parts.siblings(
            DynamicSharedCheck(
                if_shared=SharedWarningLabel(
                    rendering_parts.text(
                        f" # Repeated python obj at 0x{node_id:x}"
                    )
                ),
                node_id=node_id,
                seen_more_than_once=shared_object_tracker.seen_more_than_once,
            ),
            rendering.annotations,
        ),
    )

  return rendering



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handler for user-specified autovisualizers."""

from __future__ import annotations

from typing import Any

from treescope import lowering
from treescope import renderers
from treescope import rendering_parts
from treescope._internal import figures_impl
from treescope._internal import object_inspection
from treescope._internal.api import autovisualize

IPythonVisualization = autovisualize.IPythonVisualization
VisualizationFromTreescopePart = autovisualize.VisualizationFromTreescopePart
ChildAutovisualizer = autovisualize.ChildAutovisualizer


def use_autovisualizer_if_present(
    node: Any,
    path: str | None,
    node_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Treescope wrapper hook that runs the active autovisualizer."""
  autoviz = autovisualize.active_autovisualizer.get()
  if autoviz is None:
    return NotImplemented

  result = autoviz(node, path)
  if result is None:
    # Continue as normal.
    return NotImplemented

  elif isinstance(
      result, IPythonVisualization | VisualizationFromTreescopePart
  ):
    # We should use this visualization instead.

    # Fallback: Render normally for round-trip mode, and if this wasn't a
    # replaced object.
    with autovisualize.active_autovisualizer.set_scoped(
        lambda value, path: None
    ):
      ordinary_result = node_renderer(node, path)

    if isinstance(result, IPythonVisualization):
      display_obj = result.display_object
      if isinstance(display_obj, figures_impl.TreescopeFigure):
        ipy_rendering = display_obj.treescope_part
      elif isinstance(display_obj, object_inspection.HasReprHtml):

        def _thunk(_):
          html_rendering = object_inspection.to_html(display_obj)
          if html_rendering:
            return rendering_parts.embedded_iframe(
                embedded_html=html_rendering,
                fallback_in_text_mode=rendering_parts.abbreviation_color(
                    rendering_parts.text("<rich HTML visualization>")
                ),
            )
          else:
            return rendering_parts.error_color(
                rendering_parts.text(
                    "<Autovisualizer returned a Visualization with an invalid"
                    f" display object {display_obj}>"
                )
            )

        ipy_rendering = lowering.maybe_defer_rendering(
            _thunk,
            lambda: rendering_parts.text(
                "<rich HTML visualization loading...>"
            ),
        )
      else:
        # Bad display object
        ipy_rendering = rendering_parts.build_one_line_tree_node(
            line=rendering_parts.error_color(
                rendering_parts.text(
                    "<Autovisualizer returned a Visualization with an invalid"
                    f" display object {display_obj}>"
                )
            ),
            path=path,
        )
      if result.replace:
        replace = True
        rendering_and_annotations = (
            rendering_parts.build_custom_foldable_tree_node(
                label=rendering_parts.abbreviation_color(
                    rendering_parts.text(
                        f"<Visualization of {type(node).__name__}"
                    )
                ),
                contents=rendering_parts.siblings(
                    rendering_parts.fold_condition(
                        expanded=rendering_parts.siblings(
                            rendering_parts.abbreviation_color(
                                rendering_parts.text(":")
                            ),
                            rendering_parts.indented_children([ipy_rendering]),
                        )
                    ),
                    rendering_parts.abbreviation_color(
                        rendering_parts.text(">")
                    ),
                ),
                path=path,
                expand_state=rendering_parts.ExpandState.EXPANDED,
            )
        )
      else:
        replace = False
        internal_contents = (
            rendering_parts.floating_annotation_with_separate_focus(
                rendering_parts.fold_condition(
                    collapsed=rendering_parts.comment_color(
                        rendering_parts.text("(Visualization hidden)")
                    ),
                    expanded=ipy_rendering,
                )
            )
        )
        rendering_and_annotations = (
            rendering_parts.RenderableAndLineAnnotations(
                renderable=rendering_parts.in_outlined_box(
                    rendering_parts.build_custom_foldable_tree_node(
                        contents=internal_contents,
                        expand_state=rendering_parts.ExpandState.EXPANDED,
                    ).renderable
                ),
                annotations=rendering_parts.empty_part(),
            )
        )

    else:
      assert isinstance(result, VisualizationFromTreescopePart)
      replace = True
      rendering_and_annotations = result.rendering

    if replace:
      in_roundtrip_with_annotations = rendering_parts.siblings_with_annotations(
          ordinary_result,
          extra_annotations=[
              rendering_parts.comment_color(
                  rendering_parts.text(
                      "  # Visualization hidden in roundtrip mode"
                  )
              )
          ],
      )
      return rendering_parts.RenderableAndLineAnnotations(
          renderable=rendering_parts.roundtrip_condition(
              roundtrip=in_roundtrip_with_annotations.renderable,
              not_roundtrip=rendering_and_annotations.renderable,
          ),
          annotations=rendering_parts.roundtrip_condition(
              roundtrip=in_roundtrip_with_annotations.annotations,
              not_roundtrip=rendering_and_annotations.annotations,
          ),
      )
    else:
      return rendering_parts.siblings_with_annotations(
          ordinary_result, rendering_and_annotations
      )

  elif isinstance(result, ChildAutovisualizer):
    # Use a different autovisualizer while rendering this object.
    with autovisualize.active_autovisualizer.set_scoped(result.autovisualizer):
      return node_renderer(node, path)

  else:
    return rendering_parts.build_one_line_tree_node(
        line=rendering_parts.error_color(
            rendering_parts.text(
                f"<Autovizualizer returned an invalid value {result}; expected"
                " IPythonVisualization, VisualizationFromTreescopePart,"
                " ChildAutovisualizer, or None>"
            )
        ),
        path=path,
    )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NDArray adapter interface.

This module defines an interface for adapters that support rendering of
multi-dimensional arrays (tensors) in treescope. This can be used to add support
for a variety of array types, including numpy arrays, JAX arrays, and others.
"""

from __future__ import annotations

import abc
import dataclasses
from typing import Any, Generic, TypeVar

import numpy as np

T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class PositionalAxisInfo:
  """Marks an axis as being an ordinary (positional) axis.

  Attributes:
    axis_logical_index: The logical index of the axis. This is the index of the
      axis in the underlying array data.
    size: The size of the axis.
  """

  axis_logical_index: int
  size: int

  def logical_key(self) -> int:
    return self.axis_logical_index


@dataclasses.dataclass(frozen=True)
class NamedPositionlessAxisInfo:
  """Marks an axis as being accessible by name only.

  Attributes:
    axis_name: The name of the axis.
    size: The size of the axis.
  """

  axis_name: Any
  size: int

  def logical_key(self) -> Any:
    return self.axis_name


@dataclasses.dataclass(frozen=True)
class NamedPositionalAxisInfo:
  """Marks an axis as being accessible by name or by position.

  Attributes:
    axis_logical_index: The logical index of the axis. This is the index of the
      axis in the underlying array data.
    axis_name: The name of the axis.
    size: The size of the axis.
  """

  axis_logical_index: int
  axis_name: int
  size: int

  def logical_key(self) -> int:
    return self.axis_logical_index


@dataclasses.dataclass(frozen=True)
class ShardingInfo:
  """Summary of the sharding of an array.

  Attributes:
    shard_shape: Shape of a single shard. Should be the same length as the
      number of slices in each value of `device_index_to_shard_slices`.
    device_index_to_shard_slices: A mapping from device index to the tuple of
      per-axis indices or slices of the original array that is assigned to that
      device. Each entry of this tuple should either be an int or a slice
      object. If an int, that axis should not appear in shard_shape (e.g. the
      full array is formed by stacking the shards along a new axis). If a slice,
      the corresponding axis should appear in shard_shape, and the slice should
      be the full slice ``slice(None)`` (if the array is not sharded over this
      axis) or a slice that matches the corresponding entry in `shard_shape` (if
      the full array is formed by concatenating the shards along this axis).
    device_type: The type of device that the array is sharded across, as a
      string (e.g. "CPU", "TPU", "GPU").
    fully_replicated: Whether the array is fully replicated across all devices.
  """

  shard_shape: tuple[int, ...]
  device_index_to_shard_slices: dict[int, tuple[slice | int, ...]]
  device_type: str
  fully_replicated: bool = False


AxisInfo = (
    PositionalAxisInfo | NamedPositionlessAxisInfo | NamedPositionalAxisInfo
)


class NDArrayAdapter(abc.ABC, Generic[T]):
  """An adapter to support rendering a multi-dimensional array (tensor) type."""

  @abc.abstractmethod
  def get_axis_info_for_array_data(self, array: T) -> tuple[AxisInfo, ...]:
    """Returns axis information for each axis in the given array.

    This method should return a tuple with an AxisInfo entry for each axis in
    the array. Array axes can be one of three types:

      * Positional axes have an index and a size, and can be accessed by
        position. This is common in ordinary NDArrays.
      * Named positionless axes have a name and a size, and can be accessed by
        name only. This is how `penzai.core.named_axes` treats named axes.
      * Named positional axes have an index, a name, and a size, and can be
        accessed by either position or name. This is how PyTorch treats named
        axes.

    Note that positional axes have an explicit "logical index", which may or may
    not match their position in the underlying array data; this makes it
    possible to support "views" of underlying array data that have a different
    axis ordering than the original data. (`penzai.core.named_axes` uses this.)

    Args:
      array: The array to get axis information for.

    Returns:
      A tuple with an AxisInfo entry for each axis in the array. The ordering
      must be consistent with the ordering expected by
      `get_array_data_with_truncation`.
    """
    raise NotImplementedError(
        "Subclasses must override `get_axis_info_for_array_data`."
    )

  @abc.abstractmethod
  def get_array_data_with_truncation(
      self,
      array: T,
      mask: T | None,
      edge_items_per_axis: tuple[int | None, ...],
  ) -> tuple[np.ndarray, np.ndarray]:
    """Returns a numpy array with truncated array (and mask) data.

    This method should construct a numpy array whose contents are a truncated
    version of the given array's data; this array will be used to construct the
    actual array visualization. It is also responsible for broadcasting the mask
    appropriately and returning a compatible truncation of it.

    This method may be called many times when rendering a large structure of
    arrays (once per array), so it should be as fast as possible. We suggest
    doing truncation on an accelerator device and then copying the result, if
    possible, to avoid unnecessary data transfer.

    Args:
      array: The array to get data for.
      mask: An optional mask array provided by the user, which should be
        broadcast-compatible with ``array``. (If it is not compatible, the user
        has provided an invalid mask, and this method should raise an
        informative exception.) Can be None if no mask is provided.
      edge_items_per_axis: A tuple with one entry for each axis in the array.
        Each entry is either the number of items to keep on each side of this
        axis, or None to keep all items. The ordering will be consistent with
        the axis order returned by `get_axis_info_for_array_data`, i.e. the
        entry ``k`` in ``edge_items`` corresponds to the entry ``k`` in the axis
        info tuple, regardless of the logical indices or axis names.

    Returns:
      A tuple ``(truncated_data, truncated_mask)``. ``truncated_data`` should be
      a numpy array with a truncated version of the given array's data. If
      entry ``k`` in ``edge_items`` is ``None``, axis ``k`` should have
      the same size as the ``size`` field of the entry ``k`` returned by
      ``get_axis_info_for_array_data``. If entry ``k`` in ``edge_items``
      is not ``None``, axis ``k`` should have a size of
      ``edge_items[k] * 2 + 1``, and the middle element can be arbitrary.
      ``truncated_mask`` should be a numpy array with the same shape as
      ``truncated_data`` containing a truncated, broadcasted version of the
      mask; the middle element of the mask must be ``False`` for each truncated
      axis.
    """
    raise NotImplementedError(
        "Subclasses must override `get_array_data_with_truncation`."
    )

  @abc.abstractmethod
  def get_array_summary(self, array: T, fast: bool) -> str:
    """Summarizes the contents of the given array.

    The summary returned by this method will be used as a one-line summary of
    the array in treescope when automatically visualized.

    If the ``fast`` argument is True, the method should return a summary that
    can be computed quickly, ideally without any device computation. If it is
    False, the method can return a more detailed summary, but it should still
    be fast enough to be called many times when rendering a large structure of
    arrays.

    Args:
      array: The array to summarize.
      fast: Whether to return a fast summary that can be computed without
        expensive device computation.

    Returns:
      A summary of the given array's contents. The summary should be a single
      line of text. It will be wrapped between angle brackets (< and >) when
      rendered.
    """
    raise NotImplementedError("Subclasses must override `get_array_summary`.")

  def get_numpy_dtype(self, array: T) -> np.dtype | None:
    """Returns the numpy dtype of the given array.

    This should match the dtype of the array returned by
    `get_array_data_with_truncation`.

    Args:
      array: The array to summarize.

    Returns:
      The numpy dtype of the given array, or None if the array does not have a
      numpy dtype.
    """
    raise NotImplementedError("Subclasses must override `get_numpy_dtype`.")

  def get_sharding_info_for_array_data(self, array: T) -> ShardingInfo | None:
    """Summarizes the sharding of the given array's data.

    The summary returned by this method will be used to render a sharding for
    the array when automatic visualization is enabled.

    Args:
      array: The array to summarize.

    Returns:
      A summary of the given array's sharding, or None if it does not have a
      sharding.
    """
    # Default implementation: don't show any sharding information.
    del array

  def should_autovisualize(self, array: T) -> bool:
    """Returns True if the given array should be automatically visualized.

    If this method returns True, the array will be automatically visualized
    by the array visualizer if it is enabled.

    Args:
      array: The array to possibly visualize.

    Returns:
      True if the given array should be automatically visualized.
    """
    del array
    return True



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy setup logic for adding Numpy support to treescope."""
from __future__ import annotations

from typing import Any

import numpy as np
from treescope import canonical_aliases
from treescope import dtype_util
from treescope import lowering
from treescope import ndarray_adapters
from treescope import renderers
from treescope import rendering_parts
from treescope import type_registries


def _truncate_and_copy(
    array_source: np.ndarray,
    array_dest: np.ndarray,
    prefix_slices: tuple[slice, ...],
    remaining_edge_items_per_axis: tuple[int | None, ...],
) -> None:
  """Recursively copy values along the edges of a source into a destination.

  This function mutates the destination array in place, copying parts of input
  array into them, so that it contains a truncated versions of the original
  array.

  Args:
    array_source: Source array, which we will truncate.
    array_dest: Destination array, whose axis sizes will be either the same as
      `array_source` or of size `2 * edge_items + 1` depending on the
      truncation.
    prefix_slices: Prefix of slices for the source and destination.
    remaining_edge_items_per_axis: Number of edge items to keep for each axis,
      ignoring any axes whose slices are already computed in `source_slices`.
  """
  if not remaining_edge_items_per_axis:
    # Perform the base case slice.
    assert (
        len(prefix_slices) == len(array_source.shape) == len(array_dest.shape)
    )
    array_dest[prefix_slices] = array_source[prefix_slices]
  else:
    # Recursive step.
    axis = len(prefix_slices)
    edge_items = remaining_edge_items_per_axis[0]
    if edge_items is None:
      # Don't need to slice.
      _truncate_and_copy(
          array_source=array_source,
          array_dest=array_dest,
          prefix_slices=prefix_slices + (slice(None),),
          remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
      )
    else:
      assert array_source.shape[axis] > 2 * edge_items
      _truncate_and_copy(
          array_source=array_source,
          array_dest=array_dest,
          prefix_slices=prefix_slices + (slice(None, edge_items),),
          remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
      )
      _truncate_and_copy(
          array_source=array_source,
          array_dest=array_dest,
          prefix_slices=prefix_slices + (slice(-edge_items, None),),
          remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
      )


class NumpyArrayAdapter(ndarray_adapters.NDArrayAdapter[np.ndarray]):
  """NDArray adapter for numpy arrays."""

  def get_axis_info_for_array_data(
      self, array: np.ndarray
  ) -> tuple[ndarray_adapters.AxisInfo, ...]:
    return tuple(
        ndarray_adapters.PositionalAxisInfo(i, size)
        for i, size in enumerate(array.shape)
    )

  def get_array_data_with_truncation(
      self,
      array: np.ndarray,
      mask: np.ndarray | None,
      edge_items_per_axis: tuple[int | None, ...],
  ) -> tuple[np.ndarray, np.ndarray]:

    if mask is None:
      mask = np.ones((1,) * array.ndim, dtype=bool)

    # Broadcast mask. (Note: Broadcasting a Numpy array does not copy data.)
    mask = np.broadcast_to(mask, array.shape)

    if edge_items_per_axis == (None,) * array.ndim:
      # No truncation.
      return array, mask

    dest_shape = [
        size if edge_items is None else 2 * edge_items + 1
        for size, edge_items in zip(array.shape, edge_items_per_axis)
    ]
    array_dest = np.zeros(dest_shape, array.dtype)
    mask_dest = np.zeros(dest_shape, bool)
    _truncate_and_copy(
        array_source=array,
        array_dest=array_dest,
        prefix_slices=(),
        remaining_edge_items_per_axis=edge_items_per_axis,
    )
    _truncate_and_copy(
        array_source=mask,
        array_dest=mask_dest,
        prefix_slices=(),
        remaining_edge_items_per_axis=edge_items_per_axis,
    )
    return array_dest, mask_dest

  def get_array_summary(self, array: np.ndarray, fast: bool) -> str:
    output_parts = ["np.ndarray "]

    output_parts.append(dtype_util.get_dtype_name(array.dtype))
    output_parts.append(repr(array.shape))

    if array.size > 0 and array.size < 100_000 and not fast:
      is_floating = dtype_util.is_floating_dtype(array.dtype)
      is_integer = dtype_util.is_integer_dtype(array.dtype)
      is_bool = np.issubdtype(array.dtype, np.bool_)

      if is_floating:
        isfinite = np.isfinite(array)
        any_finite = np.any(isfinite)
        inf_to_nan = np.where(
            isfinite, array, np.array(np.nan, dtype=array.dtype)
        )
        mean = np.nanmean(inf_to_nan)
        std = np.nanstd(inf_to_nan)

        if any_finite:
          output_parts.append(f" ≈{float(mean):.2} ±{float(std):.2}")
          output_parts.append(
              f" [≥{float(np.nanmin(array)):.2}, ≤{float(np.nanmax(array)):.2}]"
          )

      if is_integer:
        output_parts.append(f" [≥{np.min(array):_d}, ≤{np.max(array):_d}]")

      if is_floating or is_integer:
        ct_zero = np.count_nonzero(array == 0)
        if ct_zero:
          output_parts.append(f" zero:{ct_zero:_d}")

        ct_nonzero = np.count_nonzero(array)
        if ct_nonzero:
          output_parts.append(f" nonzero:{ct_nonzero:_d}")

      if is_floating:
        ct_nan = np.count_nonzero(np.isnan(array))
        if ct_nan:
          output_parts.append(f" nan:{ct_nan:_d}")

        ct_inf = np.count_nonzero(np.isposinf(array))
        if ct_inf:
          output_parts.append(f" inf:{ct_inf:_d}")

        ct_neginf = np.count_nonzero(np.isneginf(array))
        if ct_neginf:
          output_parts.append(f" -inf:{ct_neginf:_d}")

      if is_bool:
        ct_true = np.count_nonzero(array)
        if ct_true:
          output_parts.append(f" true:{ct_true:_d}")

        ct_false = np.count_nonzero(np.logical_not(array))
        if ct_false:
          output_parts.append(f" false:{ct_false:_d}")

    return "".join(output_parts)

  def get_numpy_dtype(self, array: np.ndarray) -> np.dtype:
    return array.dtype


def render_ndarrays(
    node: np.ndarray,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a numpy array."""
  del subtree_renderer
  assert isinstance(node, np.ndarray)
  adapter = NumpyArrayAdapter()

  def _placeholder() -> rendering_parts.RenderableTreePart:
    return rendering_parts.deferred_placeholder_style(
        rendering_parts.text(adapter.get_array_summary(node, fast=True))
    )

  def _thunk(placeholder_expand_state: rendering_parts.ExpandState | None):
    # Is this array simple enough to render without a summary?
    node_repr = repr(node)
    if "\n" not in node_repr and "..." not in node_repr:
      rendering = rendering_parts.text(f"np.{node_repr}")
    else:
      if node_repr.count("\n") <= 15:
        if placeholder_expand_state is None:
          default_expand_state = rendering_parts.ExpandState.WEAKLY_EXPANDED
        else:
          default_expand_state = placeholder_expand_state
      else:
        # Always start big NDArrays in collapsed mode to hide irrelevant detail.
        default_expand_state = rendering_parts.ExpandState.COLLAPSED

      # Render it with a summary.
      summarized = adapter.get_array_summary(node, fast=False)
      rendering = rendering_parts.build_custom_foldable_tree_node(
          label=rendering_parts.abbreviation_color(
              rendering_parts.comment_color_when_expanded(
                  rendering_parts.siblings(
                      rendering_parts.fold_condition(
                          expanded=rendering_parts.text("# "),
                          collapsed=rendering_parts.text("<"),
                      ),
                      summarized,
                      rendering_parts.fold_condition(
                          collapsed=rendering_parts.text(">")
                      ),
                  )
              )
          ),
          contents=rendering_parts.fold_condition(
              expanded=rendering_parts.indented_children(
                  [rendering_parts.text(node_repr)]
              )
          ),
          path=path,
          expand_state=default_expand_state,
      ).renderable

    return rendering

  return rendering_parts.RenderableAndLineAnnotations(
      renderable=lowering.maybe_defer_rendering(
          main_thunk=_thunk, placeholder_thunk=_placeholder
      ),
      annotations=rendering_parts.build_copy_button(path),
  )


def render_dtype_instances(
    node: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a np.dtype, adding the `np.` qualifier."""
  del subtree_renderer
  if not isinstance(node, np.dtype):
    return NotImplemented

  dtype_name = node.name
  if dtype_name in np.sctypeDict and node is np.dtype(
      np.sctypeDict[dtype_name]
  ):
    # Use the named type. (Sometimes extended dtypes don't print in a
    # roundtrippable way otherwise.)
    dtype_string = f"dtype({repr(dtype_name)})"
  else:
    # Hope that `repr` is already round-trippable (true for builtin numpy types)
    # and add the "numpy." prefix as needed.
    dtype_string = repr(node)

  return rendering_parts.build_one_line_tree_node(
      line=rendering_parts.siblings(
          rendering_parts.roundtrip_condition(
              roundtrip=rendering_parts.text("np.")
          ),
          dtype_string,
      ),
      path=path,
  )


def set_up_treescope():
  """Sets up treescope to render Numpy objects."""
  type_registries.NDARRAY_ADAPTER_REGISTRY[np.ndarray] = NumpyArrayAdapter()
  type_registries.TREESCOPE_HANDLER_REGISTRY[np.ndarray] = render_ndarrays
  type_registries.TREESCOPE_HANDLER_REGISTRY[np.dtype] = render_dtype_instances

  canonical_aliases.populate_from_public_api(
      np, canonical_aliases.prefix_filter("numpy", excludes=("numpy.core",))
  )
